# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the DC/DC Converter.
"""
import numpy as np
import schedula as sh

dsp = sh.BlueDispatcher(
    name='DC/DC Converter',
    description='Models the DC/DC Converter.'
)


@sh.add_function(
    dsp, inputs_kwargs=True, inputs_defaults=True,
    outputs=['dcdc_converter_electric_powers_demand']
)
def calculate_dcdc_converter_electric_powers_demand(
        dcdc_converter_electric_powers, dcdc_converter_efficiency=.95):
    """
    Calculate DC/DC converter electric power demand [kW].

    :param dcdc_converter_electric_powers:
        DC/DC converter electric power [kW].
    :type dcdc_converter_electric_powers: numpy.array | float

    :param dcdc_converter_efficiency:
        DC/DC converter efficiency [-].
    :type dcdc_converter_efficiency: float

    :return:
        DC/DC converter electric power demand [kW].
    :rtype: numpy.array | float
    """
    from ..motors.p4 import calculate_motor_p4_powers_v1 as func
    return func(dcdc_converter_electric_powers, dcdc_converter_efficiency)


@sh.add_function(dsp, outputs=['dcdc_converter_electric_powers'])
def calculate_dcdc_converter_electric_powers_v1(
        dcdc_converter_electric_powers_demand, dcdc_converter_efficiency):
    """
    Calculate DC/DC converter electric power [kW].

    :param dcdc_converter_electric_powers_demand:
        DC/DC converter electric power demand [kW].
    :type dcdc_converter_electric_powers_demand: numpy.array | float

    :param dcdc_converter_efficiency:
        DC/DC converter efficiency [-].
    :type dcdc_converter_efficiency: float

    :return:
        DC/DC converter electric power [kW].
    :rtype: numpy.array | float
    """
    from ..motors.p4 import calculate_motor_p4_electric_powers as f
    return f(dcdc_converter_electric_powers_demand, dcdc_converter_efficiency)


@sh.add_function(dsp, outputs=['dcdc_converter_electric_powers'])
def calculate_dcdc_converter_electric_powers(
        dcdc_converter_currents, service_battery_nominal_voltage):
    """
    Calculates DC/DC converter electric powers [kW].

    :param dcdc_converter_currents:
        DC/DC converter currents [A].
    :type dcdc_converter_currents: numpy.array | float

    :param service_battery_nominal_voltage:
        Service battery nominal voltage [V].
    :type service_battery_nominal_voltage: float

    :return:
        DC/DC converter electric power [kW].
    :rtype: numpy.array | float
    """
    return dcdc_converter_currents * (service_battery_nominal_voltage / 1000)


@sh.add_function(dsp, outputs=['dcdc_converter_currents'])
def calculate_dcdc_converter_currents(
        dcdc_converter_electric_powers, service_battery_nominal_voltage):
    """
    Calculate DC/DC converter current [A].

    :param dcdc_converter_electric_powers:
        DC/DC converter electric power [kW].
    :type dcdc_converter_electric_powers: numpy.array | float

    :param service_battery_nominal_voltage:
        Service battery nominal voltage [V].
    :type service_battery_nominal_voltage: float

    :return:
        DC/DC converter currents [A].
    :rtype: numpy.array | float
    """
    c = 1000 / service_battery_nominal_voltage
    return dcdc_converter_electric_powers * c


@sh.add_function(dsp, outputs=['dcdc_current_model'])
def define_dcdc_current_model(dcdc_charging_currents):
    """
    Defines an DC/DC converter current model that predicts its current [A].

    :param dcdc_charging_currents:
        Mean charging currents of the DC/DC converter (for negative and positive
        power)[A].
    :type dcdc_charging_currents: (float, float)

    :return:
        DC/DC converter current model.
    :rtype: callable
    """
    from ..motors.alternator.current import define_alternator_current_model
    return define_alternator_current_model(dcdc_charging_currents)


@sh.add_function(dsp, outputs=['dcdc_current_model'])
def calibrate_dcdc_current_model(
        dcdc_converter_currents, times, service_battery_state_of_charges,
        service_battery_charging_statuses, service_battery_initialization_time):
    """
    Calibrates an DC/DC current model that predicts DC/DC current [A].

    :param dcdc_converter_currents:
        DC/DC converter currents [A].
    :type dcdc_converter_currents: numpy.array

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param service_battery_state_of_charges:
        State of charge of the service battery [%].
    :type service_battery_state_of_charges: numpy.array

    :param service_battery_charging_statuses:
        Service battery charging statuses (0: Discharge, 1: Charging, 2: BERS,
        3: Initialization) [-].
    :type service_battery_charging_statuses: numpy.array

    :param service_battery_initialization_time:
        Service battery initialization time delta [s].
    :type service_battery_initialization_time: float

    :return:
        DC/DC converter current model.
    :rtype: callable
    """
    from ..motors.alternator.current import AlternatorCurrentModel
    statuses = service_battery_charging_statuses
    return AlternatorCurrentModel().fit(
        dcdc_converter_currents, np.in1d(statuses, (1, 3)), times,
        service_battery_state_of_charges, statuses,
        init_time=service_battery_initialization_time
    )


@sh.add_function(dsp, outputs=['dcdc_converter_currents'])
def predict_dcdc_converter_currents(
        dcdc_current_model, times, service_battery_state_of_charges,
        service_battery_charging_statuses, service_battery_initialization_time):
    """
    Predict DC/DC converter current [A].

    :param dcdc_current_model:
        DC/DC converter current model.
    :type dcdc_current_model: callable

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param service_battery_state_of_charges:
        State of charge of the service battery [%].
    :type service_battery_state_of_charges: numpy.array

    :param service_battery_charging_statuses:
        Service battery charging statuses (0: Discharge, 1: Charging, 2: BERS,
        3: Initialization) [-].
    :type service_battery_charging_statuses: numpy.array

    :param service_battery_initialization_time:
        Service battery initialization time delta [s].
    :type service_battery_initialization_time: float

    :return:
        DC/DC converter currents [A].
    :rtype: numpy.array
    """
    return dcdc_current_model.predict(np.column_stack((
        times, service_battery_state_of_charges,
        service_battery_charging_statuses
    )), init_time=service_battery_initialization_time)


@sh.add_function(dsp, outputs=['dcdc_converter_currents'], weight=sh.inf(10, 3))
def default_dcdc_converter_currents(times, is_hybrid):
    """
    Return zero current if the vehicle is not hybrid [A].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param is_hybrid:
        Is the vehicle hybrid?
    :type is_hybrid: bool

    :return:
        DC/DC converter currents [A].
    :rtype: numpy.array
    """
    from ..motors.p4 import default_motor_p4_powers as func
    return func(times, is_hybrid)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to define theoretical times, velocities, and gears.

Sub-Modules:

.. currentmodule:: co2mpas.core.model.physical.cycle

.. autosummary::
    :nosignatures:
    :toctree: cycle/

    NEDC
    WLTP
"""
import numpy as np
import schedula as sh
import co2mpas.utils as co2_utl
from co2mpas.defaults import dfl
from .NEDC import dsp as _nedc_cycle, is_manual
from .WLTP import dsp as _wltp_cycle

dsp = sh.BlueDispatcher(
    name='Cycle model',
    description='Returns the theoretical times, velocities, and gears.'
)
dsp.add_data('time_sample_frequency', dfl.values.time_sample_frequency)


# noinspection PyMissingOrEmptyDocstring
def is_nedc(kw):
    return kw.get('cycle_type') == 'NEDC'


# noinspection PyMissingOrEmptyDocstring
def is_wltp(kw):
    return kw.get('cycle_type') == 'WLTP'


dsp.add_dispatcher(
    include_defaults=True,
    dsp=_nedc_cycle,
    inputs=(
        'gear_box_type', 'gears', 'k1', 'k2', 'k5', 'max_gear', 'times',
        {'cycle_type': sh.SINK}
    ),
    outputs=('gears', 'initial_temperature', 'max_time', 'velocities'),
    input_domain=is_nedc
)

dsp.add_dispatcher(
    include_defaults=True,
    dsp=_wltp_cycle,
    inputs=(
        'accelerations', 'downscale_factor', 'motive_powers',
        'downscale_factor_threshold', 'downscale_phases', 'engine_max_power',
        'engine_speed_at_max_power', 'full_load_curve', 'gear_box_type',
        'gears', 'idle_engine_speed', 'speed_velocity_ratios', 'max_velocity',
        'max_speed_velocity_ratio', 'inertial_factor', 'road_loads', 'times',
        'unladen_mass', 'vehicle_mass', 'velocities', 'wltp_base_model',
        'engine_max_speed', 'wltp_class', {'cycle_type': sh.SINK}
    ),
    outputs=(
        'theoretical_motive_powers', 'gears', 'initial_temperature', 'max_time',
        'velocities', 'theoretical_velocities'
    ),
    input_domain=is_wltp
)


@sh.add_function(dsp, outputs=['time_length'])
def calculate_time_length(time_sample_frequency, max_time):
    """
    Returns the length of the time vector [-].

    :param time_sample_frequency:
        Time frequency [1/s].
    :type time_sample_frequency: float

    :param max_time:
        Maximum time [s].
    :type max_time: float

    :return:
        length of the time vector [-].
    :rtype: int
    """
    return np.floor(max_time * time_sample_frequency) + 1


@sh.add_function(dsp, outputs=['times'])
def cycle_times(time_sample_frequency, time_length):
    """
    Returns the time vector with constant time step [s].

    :param time_sample_frequency:
        Time frequency [1/s].
    :type time_sample_frequency: float

    :param time_length:
        Length of the time vector [-].
    :type time_length: float

    :return:
        Time vector [s].
    :rtype: numpy.array
    """

    dt = 1 / time_sample_frequency

    return np.arange(0.0, time_length, dtype=float) * dt


dsp.add_function(function=len, inputs=['velocities'], outputs=['time_length'])
dsp.add_function(
    function=len, inputs=['gears'], outputs=['time_length'], weight=1
)


def _extract_indices(bag_phases):
    pit, bag_phases = [], np.asarray(bag_phases)
    n = len(bag_phases) - 1
    for bf in np.unique(bag_phases):
        i = np.where(bf == bag_phases)[0]
        pit.append((i.min(), min(i.max() + 1, n)))
    return sorted(pit)


@sh.add_function(dsp, outputs=['phases_integration_times'])
def extract_phases_integration_times(times, bag_phases):
    """
    Extracts the cycle phases integration times [s] from bag phases vector.

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param bag_phases:
        Bag phases [-].
    :type bag_phases: numpy.array

    :return:
        Cycle phases integration times [s].
    :rtype: tuple
    """

    return tuple((times[i], times[j]) for i, j in _extract_indices(bag_phases))


@sh.add_function(dsp, outputs=['phases_integration_times'], weight=10)
def select_phases_integration_times(cycle_type):
    """
    Selects the cycle phases integration times [s].

    :param cycle_type:
        Cycle type (WLTP or NEDC).
    :type cycle_type: str

    :return:
        Cycle phases integration times [s].
    :rtype: tuple
    """

    from co2mpas.defaults import dfl
    v = dfl.functions.select_phases_integration_times.INTEGRATION_TIMES
    return tuple(co2_utl.pairwise(v[cycle_type.upper()]))



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the CMV Approach.
"""
import itertools
import collections
import numpy as np
import schedula as sh
import co2mpas.utils as co2_utl
from co2mpas.defaults import dfl
from .core import prediction_gears_gsm, define_gear_filter

dsp = sh.BlueDispatcher(name='Corrected Matrix Velocity Approach')

dsp.add_data('stop_velocity', dfl.values.stop_velocity)


def _correct_gsv(gsv, stop_velocity):
    """
    Corrects gear shifting velocity matrix from unreliable limits.

    :param gsv:
        Gear shifting velocity matrix.
    :type gsv: dict

    :param stop_velocity:
        Maximum velocity to consider the vehicle stopped [km/h].
    :type stop_velocity: float

    :return:
        Gear shifting velocity matrix corrected from unreliable limits.
    :rtype: dict
    """

    gsv[0] = [0, (stop_velocity, (dfl.INF, 0))]

    # noinspection PyMissingOrEmptyDocstring
    def func(x):
        return not x and float('inf') or 1 / x

    for v0, v1 in co2_utl.pairwise(gsv.values()):
        up0, s0, down1, s1 = v0[1][0], v0[1][1][1], v1[0][0], v1[0][1][1]

        if down1 + s1 <= v0[0]:
            v0[1], v1[0] = up0 + s0, up0 - s0
        elif up0 >= down1:
            v0[1], v1[0] = up0 + s0, down1 - s1
            continue
        elif (v0[1][1][0], func(s0)) >= (v1[0][1][0], func(s1)):
            v0[1], v1[0] = up0 + s0, up0 - s0
        else:
            v0[1], v1[0] = down1 + s1, down1 - s1

        v0[1] += stop_velocity

    gsv[max(gsv)][1] = dfl.INF

    return gsv


def _filter_gear_shifting_velocity(limits, stop_velocity):
    def _rjt_out(x, default):
        if x:
            x = np.asarray(x)

            # noinspection PyTypeChecker
            m, (n, s) = np.median(x), (len(x), np.std(x))

            y = s and 2 > (abs(x - m) / s)

            if s and y.any():
                y = x[y]

                # noinspection PyTypeChecker
                m, (n, s) = np.median(y), (len(y), np.std(y))

            return m, (n, s)
        else:
            return default

    max_gear = max(limits)
    gsv = collections.OrderedDict()
    for k in range(max_gear + 1):
        v0, v1 = limits.get(k, [[], []])
        gsv[k] = [_rjt_out(v0, (-1, (0, 0))), _rjt_out(v1, (dfl.INF, (0, 0)))]

    return _correct_gsv(gsv, stop_velocity)


def _identify_gear_shifting_velocity_limits(gears, velocities, stop_velocity):
    """
    Identifies gear shifting velocity matrix.

    :param gears:
        Gear vector [-].
    :type gears: numpy.array

    :param velocities:
        Vehicle velocity [km/h].
    :type velocities: numpy.array

    :param stop_velocity:
        Maximum velocity to consider the vehicle stopped [km/h].
    :type stop_velocity: float

    :return:
        Gear shifting velocity matrix.
    :rtype: dict
    """

    limits = {}

    for v, (g0, g1) in zip(velocities, co2_utl.pairwise(gears)):
        if v >= stop_velocity and g0 != g1:
            limits[g0] = limits.get(g0, [[], []])
            limits[g0][int(g0 < g1)].append(v)
    return _filter_gear_shifting_velocity(limits, stop_velocity)


def _interp(r, y, nr):
    from scipy.interpolate import InterpolatedUnivariateSpline as Spline
    n = len(r)
    if n == 0:
        return np.nan * nr
    elif n == 1:
        return (y / r) * nr
    else:
        return Spline(r, y / r, k=1)(nr) * nr


# noinspection PyPep8Naming
def _convert_limits(lu, vsr, n_vsr, stop_velocity=dfl.values.stop_velocity):
    _r, _l, _u = np.array(
        sorted([vsr.get(k, 0)] + list(v) for k, v in lu.items())
    ).T
    nk, nr = np.array(sorted(sh.combine_dicts(n_vsr, base={0: 0}).items())).T
    klu = sorted(zip(
        nk.astype(int),
        _interp(_r[2:], _l[2:], nr),
        _interp(_r[1:-1], _u[1:-1], nr)
    ))
    res, n0, n1 = {}, min(k for k in n_vsr if k > 0), max(n_vsr)
    for k, l, u in klu:
        if k == 0:
            l, u = list(lu.get(k, [0, stop_velocity + 1]))
        else:
            if k == n0:
                l = lu.get(k, (stop_velocity,))[0]
            if k == n1:
                u = dfl.INF
        res[k] = [l, u]
    return res


def _grouper(iterable, n):
    """
    Collect data into fixed-length chunks or blocks.

    :param iterable:
        Iterable object.
    :param iterable: iter

    :param n:
        Length chunks or blocks.
    :type n: int
    """
    args = [iter(iterable)] * n
    return zip(*args)


# noinspection PyMissingOrEmptyDocstring,PyUnusedLocal
class CMV(collections.OrderedDict):
    def __init__(self, *args, velocity_speed_ratios=None):
        super(CMV, self).__init__(*args)
        if args and isinstance(args[0], CMV):
            if velocity_speed_ratios:
                self.convert(velocity_speed_ratios)
            else:
                velocity_speed_ratios = args[0].velocity_speed_ratios

        self.velocity_speed_ratios = velocity_speed_ratios or {}

    def __repr__(self):
        from pprint import pformat
        name = self.__class__.__name__
        items = [(k, v) for k, v in self.items()]
        vsr = pformat(self.velocity_speed_ratios)
        s = '{}({}, velocity_speed_ratios={})'.format(name, items, vsr)
        return s.replace('inf', "float('inf')")

    def fit(self, correct_gear, gears, engine_speeds_out, times, velocities,
            accelerations, motive_powers, velocity_speed_ratios, stop_velocity):
        from ..mechanical import calculate_gear_box_speeds_in
        self.clear()
        self.velocity_speed_ratios = velocity_speed_ratios
        self.update(_identify_gear_shifting_velocity_limits(
            gears, velocities, stop_velocity
        ))
        if dfl.functions.CMV.ENABLE_OPT_LOOP:
            from co2mpas.utils import mae
            gear_id, velocity_limits = zip(*sorted(self.items())[1:])
            max_gear, _inf = gear_id[-1], float('inf')
            update, predict = self.update, self.predict

            def _update_gvs(vel_limits):
                self[0] = (0, vel_limits[0])
                self[max_gear] = (vel_limits[-1], _inf)
                update(dict(zip(gear_id, _grouper(vel_limits[1:-1], 2))))

            def _error_fun(vel_limits):
                _update_gvs(vel_limits)

                g_pre = predict(
                    times, velocities, accelerations, motive_powers,
                    correct_gear=correct_gear
                )

                speed_pred = calculate_gear_box_speeds_in(
                    g_pre, velocities, velocity_speed_ratios, stop_velocity
                )
                return np.float32(mae(speed_pred, engine_speeds_out))

            x0 = [self[0][1]].__add__(
                list(itertools.chain(*velocity_limits))[:-1]
            )
            from scipy.optimize import fmin
            _update_gvs(fmin(_error_fun, x0, disp=False))

        return self

    def correct_constant_velocity(
            self, up_cns_vel=(), up_window=0.0, up_delta=0.0, dn_cns_vel=(),
            dn_window=0.0, dn_delta=0.0):
        """
        Corrects the gear shifting matrix velocity for constant velocities.

        :param up_cns_vel:
            Constant velocities to correct the upper limits [km/h].
        :type up_cns_vel: tuple[float]

        :param up_window:
            Window to identify if the shifting matrix has limits close to
            `up_cns_vel` [km/h].
        :type up_window: float

        :param up_delta:
            Delta to add to the limit if this is close to `up_cns_vel` [km/h].
        :type up_delta: float

        :param dn_cns_vel:
            Constant velocities to correct the bottom limits [km/h].
        :type dn_cns_vel: tuple[float]

        :param dn_window:
            Window to identify if the shifting matrix has limits close to
            `dn_cns_vel` [km/h].
        :type dn_window: float

        :param dn_delta:
            Delta to add to the limit if this is close to `dn_cns_vel` [km/h].
        :type dn_delta: float

        :return:
            A gear shifting velocity matrix corrected from NEDC velocities.
        :rtype: dict
        """

        def _set_velocity(velocity, const_steps, window, delta):
            for s in const_steps:
                if s < velocity < s + window:
                    return s + delta
            return velocity

        for k, v in sorted(self.items()):
            v = [
                _set_velocity(v[0], dn_cns_vel, dn_window, dn_delta),
                _set_velocity(v[1], up_cns_vel, up_window, up_delta)
            ]

            if v[0] >= v[1]:
                v[0] = v[1] + dn_delta

            try:
                if self[k - 1][1] <= v[0]:
                    v[0] = self[k - 1][1] + up_delta
            except KeyError:
                pass
            self[k] = tuple(v)

        return self

    def fig(self):
        import itertools
        import plotly.graph_objs as go
        from plotly.colors import DEFAULT_PLOTLY_COLORS
        colors = itertools.cycle(DEFAULT_PLOTLY_COLORS)
        fig = go.Figure()
        for k, v in self.items():
            color = next(colors)
            for (s, l), x in zip((('down', 'dash'), ('up', None)), v):
                if x < dfl.INF:
                    fig.add_trace(go.Scatter(
                        name='Gear %d:%s-shift' % (k, s), x=[x] * 2, y=[0, 1],
                        line=dict(color=color, dash=l), mode='lines'
                    ))
        fig.update_layout(
            title=self.__class__.__name__,
            xaxis_title="Velocity [km/h]"
        )
        return fig

    def _init_gear(self, times, velocities, accelerations, motive_powers,
                   engine_coolant_temperatures):

        keys = sorted(self.keys())
        matrix, c = {}, len(keys) - 1
        from co2mpas.utils import List
        if isinstance(velocities, List):
            for i, k in enumerate(keys):
                matrix[k] = self[k], (keys[max(0, i - 1)], keys[min(i + 1, c)])

            def _next(gear, index):
                v = velocities[index]
                (_down, _up), (g0, g1) = matrix[gear]
                if v >= _up:
                    return g1
                if v < _down:
                    return g0
                return gear
        else:
            r = velocities.shape[0]
            for i, g in enumerate(keys):
                down, up = self[g]
                matrix[g] = p = np.tile(g, r)
                p[velocities < down] = keys[max(0, i - 1)]
                p[velocities >= up] = keys[min(i + 1, c)]

            def _next(gear, index):
                return matrix[gear][index]
        return _next

    def predict(self, times, velocities, accelerations, motive_powers=None,
                engine_coolant_temperatures=None,
                correct_gear=lambda g, *args: g,
                gear_filter=define_gear_filter()):
        gears = np.zeros_like(times, int)
        gen = self.init_gear(
            gears, times, velocities, accelerations, motive_powers,
            engine_coolant_temperatures, correct_gear
        )
        for i in range(times.shape[0]):
            gears[i] = gen(i)

        # if gear_filter is not None:
        #    gears[index:times.shape[0]] = gear_filter(times, gears)

        return gears

    def init_gear(self, gears, times, velocities, accelerations,
                  motive_powers=None, engine_coolant_temperatures=None,
                  correct_gear=lambda g, *args: g):

        next_g = self._init_gear(
            times, velocities, accelerations, motive_powers,
            engine_coolant_temperatures
        )

        valid_gears = np.array(list(getattr(self, 'gears', self)))

        def get_valid_gear(g):
            if g in valid_gears:
                return g
            return valid_gears[np.abs(np.subtract(valid_gears, g)).argmin()]

        args = (
            gears, times, velocities, accelerations, motive_powers,
            engine_coolant_temperatures, next_g
        )

        def _next(i):
            if i != 0:
                g = gears[i - 1]
                return get_valid_gear(correct_gear(next_g(g, i), i, *args))
            g = valid_gears.min()
            for _ in range(len(valid_gears) + 1):
                g0 = g
                g = get_valid_gear(correct_gear(next_g(g, i), i, *args))
                if g == g0:
                    break
            return g

        return _next

    def init_speed(self, stop_velocity, gears, velocities, *args, **kwargs):
        vsr = self.velocity_speed_ratios.get

        def _next(i):
            v = velocities[i]
            r = v > stop_velocity and vsr(gears[i], 0)
            return v / r if r else 0

        return _next

    # noinspection PyPep8Naming
    def convert(self, velocity_speed_ratios):
        if velocity_speed_ratios != self.velocity_speed_ratios:
            vsr, n_vsr = self.velocity_speed_ratios, velocity_speed_ratios
            new_self = _convert_limits(self, vsr, n_vsr)
            reverse = np.diff(list(self))[0] < 0
            self.clear()
            for k, lu in sorted(new_self.items(), reverse=reverse):
                self[k] = lu

            self.velocity_speed_ratios = n_vsr

        return self


@sh.add_function(dsp, outputs=['CMV'])
def calibrate_gear_shifting_cmv(
        correct_gear, gears, engine_speeds_out, times, velocities,
        accelerations, motive_powers, velocity_speed_ratios, stop_velocity):
    """
    Calibrates a corrected matrix velocity to predict gears.

    :param correct_gear:
        A function to correct the predicted gear.
    :type correct_gear: callable

    :param gears:
        Gear vector [-].
    :type gears: numpy.array

    :param engine_speeds_out:
        Engine speed vector [RPM].
    :type engine_speeds_out: numpy.array

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param velocities:
        Vehicle velocity [km/h].
    :type velocities: numpy.array

    :param accelerations:
        Vehicle acceleration [m/s2].
    :type accelerations: numpy.array

    :param motive_powers:
        Motive power [kW].
    :type motive_powers: numpy.array

    :param velocity_speed_ratios:
        Constant velocity speed ratios of the gear box [km/(h*RPM)].
    :type velocity_speed_ratios: dict[int | float]

    :param stop_velocity:
        Maximum velocity to consider the vehicle stopped [km/h].
    :type stop_velocity: float

    :returns:
        A corrected matrix velocity to predict gears.
    :rtype: dict
    """

    cmv = CMV().fit(
        correct_gear, gears, engine_speeds_out, times, velocities,
        accelerations, motive_powers, velocity_speed_ratios, stop_velocity
    )

    return cmv


# predict gears with corrected matrix velocity
dsp.add_function(
    function=prediction_gears_gsm,
    inputs=['correct_gear', 'gear_filter', 'CMV', 'times', 'velocities',
            'accelerations', 'motive_powers', 'cycle_type',
            'velocity_speed_ratios'],
    outputs=['gears']
)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the mechanic of the vehicle.
"""
import numpy as np
import schedula as sh
from co2mpas.defaults import dfl

dsp = sh.BlueDispatcher(
    name='Vehicle free body diagram',
    description='Calculates forces and power acting on the vehicle.'
)


@sh.add_function(dsp, outputs=['velocities'])
def calculate_velocities(times, obd_velocities):
    """
    Filters the obd velocities [km/h].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param obd_velocities:
        OBD velocity vector [km/h].
    :type obd_velocities: numpy.array

    :return:
        Velocity vector [km/h].
    :rtype: numpy.array
    """
    from co2mpas.utils import median_filter
    dt_window = dfl.functions.calculate_velocities.dt_window
    return median_filter(times, obd_velocities, dt_window, np.mean)


def _integral(x, y, y0=.0):
    from scipy.interpolate import InterpolatedUnivariateSpline as Spl
    return Spl(x, y).antiderivative()(x) + y0


@sh.add_function(dsp, outputs=['distances'])
def calculate_distances(times, velocities):
    """
    Calculates the cumulative distance vector [m].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param velocities:
        Velocity vector [km/h].
    :type velocities: numpy.array

    :return:
        Cumulative distance vector [m].
    :rtype: numpy.array
    """
    d = _integral(times, velocities / 3.6, 0)
    d[1:] = np.cumsum(np.maximum(0, np.diff(d))) + d[0]
    return d


@sh.add_function(
    dsp, inputs_kwargs=True, inputs_defaults=True, outputs=['velocities']
)
def calculate_velocities_v1(times, accelerations, initial_velocity=.0):
    """
    Calculates the velocity from acceleration time series [km/h].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param accelerations:
        Acceleration vector [m/s2].
    :type accelerations: numpy.array

    :param initial_velocity:
        Initial velocity [km/h].
    :type initial_velocity: float

    :return:
        Velocity vector [km/h].
    :rtype: numpy.array
    """
    vel = _integral(times, accelerations, initial_velocity / 3.6) * 3.6
    return np.maximum(0, vel)


@sh.add_function(dsp, outputs=['accelerations'])
def calculate_accelerations(times, velocities):
    """
    Calculates the acceleration from velocity time series [m/s2].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param velocities:
        Velocity vector [km/h].
    :type velocities: numpy.array

    :return:
        Acceleration vector [m/s2].
    :rtype: numpy.array
    """
    from scipy.interpolate import InterpolatedUnivariateSpline as Spl
    acc = Spl(times, velocities / 3.6).derivative()(times)
    b = (velocities[:-1] == 0) & (velocities[1:] == velocities[:-1])
    acc[:-1][b] = 0
    if b[-1]:
        acc[-1] = 0
    return acc


@sh.add_function(dsp, outputs=['aerodynamic_resistances'])
def calculate_aerodynamic_resistances(f2, velocities):
    """
    Calculates the aerodynamic resistances of the vehicle [N].

    :param f2:
        As used in the dyno and defined by respective guidelines [N/(km/h)^2].
    :type f2: float

    :param velocities:
        Velocity vector [km/h].
    :type velocities: numpy.array | float

    :return:
        Aerodynamic resistance vector [N].
    :rtype: numpy.array | float
    """

    return f2 * velocities ** 2


dsp.add_data('n_passengers', dfl.values.n_passengers)
dsp.add_data('passenger_mass', dfl.values.passenger_mass)
dsp.add_data('cargo_mass', dfl.values.cargo_mass)
dsp.add_data('fuel_mass', dfl.values.fuel_mass)


@sh.add_function(dsp, outputs=['passengers_mass'])
def calculate_passengers_mass(n_passengers, passenger_mass):
    """
    Calculate passengers mass including driver [kg].

    :param n_passengers:
        Number of passengers including driver [-].
    :type n_passengers: int

    :param passenger_mass:
        Average passenger mass [kg].
    :type passenger_mass: float

    :return:
        Passengers mass including the driver [kg].
    :rtype: float
    """

    return passenger_mass * n_passengers


@sh.add_function(dsp, outputs=['unladen_mass'])
def calculate_unladen_mass(curb_mass, fuel_mass):
    """
    Calculate unladen mass [kg].

    :param curb_mass:
        Curb mass [kg].
    :type curb_mass: float

    :param fuel_mass:
        Fuel mass [kg].
    :type fuel_mass: float

    :return:
        Unladen mass [kg].
    :rtype: float
    """

    return curb_mass + fuel_mass


@sh.add_function(dsp, outputs=['unladen_mass'])
def calculate_unladen_mass_v1(vehicle_mass, passengers_mass, cargo_mass):
    """
    Calculate unladen mass [kg].

    :param vehicle_mass:
        Vehicle mass [kg].
    :type vehicle_mass: float

    :param passengers_mass:
        Passengers mass including the driver [kg].
    :type passengers_mass: float

    :param cargo_mass:
        Cargo mass [kg].
    :type cargo_mass: float

    :return:
        Unladen mass [kg].
    :rtype: float
    """

    return vehicle_mass - passengers_mass - cargo_mass


@sh.add_function(dsp, outputs=['curb_mass'])
def calculate_curb_mass(unladen_mass, fuel_mass):
    """
    Calculate unladen mass [kg].

    :param unladen_mass:
        Unladen mass [kg].
    :type unladen_mass: float

    :param fuel_mass:
        Fuel mass [kg].
    :type fuel_mass: float

    :return:
        Curb mass [kg].
    :rtype: float
    """

    return unladen_mass - fuel_mass


@sh.add_function(
    dsp, outputs=['vehicle_mass'], input_domain=lambda n, *a: n == 1
)
def calculate_vehicle_mass_v1(
        n_dyno_axes, test_mass, vehicle_mass_running_order):
    """
    Calculate vehicle mass [kg].

    :param test_mass:
        Test mass according to WLTP regulation 2017/1151 [kg].
    :type test_mass: float

    :param vehicle_mass_running_order:
        Mass of the vehicle in running order [kg].
    :type vehicle_mass_running_order: float

    :param n_dyno_axes:
        Number of dyno axes [-].
    :type n_dyno_axes: int

    :return:
        Vehicle mass [kg].
    :rtype: float
    """
    return test_mass + (vehicle_mass_running_order + 25) * .015


dsp.add_function(
    function_id='calculate_vehicle_mass_v2',
    function=sh.add_args(sh.bypass),
    inputs=['n_dyno_axes', 'test_mass'],
    outputs=['vehicle_mass'],
    input_domain=lambda n, *a: n != 1
)


@sh.add_function(dsp, outputs=['test_mass'])
def calculate_test_mass(
        vehicle_mass, vehicle_mass_running_order, n_dyno_axes):
    """
    Calculate test mass according to WLTP regulation 2017/1151 [kg]

    :param vehicle_mass:
         Vehicle mass [kg].
    :type vehicle_mass: float

    :param vehicle_mass_running_order:
        Mass of the vehicle in running order [kg].
    :type vehicle_mass_running_order: float

    :param n_dyno_axes:
        Number of dyno axes [-].
    :type n_dyno_axes: int

    :return:
        Test mass according to WLTP regulation 2017/1151 [kg].
    :rtype: float
    """
    if n_dyno_axes == 1:
        return vehicle_mass - (vehicle_mass_running_order + 25) * .015
    return vehicle_mass


@sh.add_function(dsp, outputs=['vehicle_mass'])
def calculate_vehicle_mass(unladen_mass, passengers_mass, cargo_mass):
    """
    Calculate vehicle mass [kg].

    :param unladen_mass:
        Unladen mass [kg].
    :type unladen_mass: float

    :param passengers_mass:
        Passengers mass including the driver [kg].
    :type passengers_mass: float

    :param cargo_mass:
        Cargo mass [kg].
    :type cargo_mass: float

    :return:
        Vehicle mass [kg].
    :rtype: float
    """

    return unladen_mass + passengers_mass + cargo_mass


@sh.add_function(dsp, outputs=['raw_frontal_area'])
def calculate_raw_frontal_area(vehicle_height, vehicle_width):
    """
    Calculates raw frontal area of the vehicle [m2].

    :param vehicle_height:
        Vehicle height [m].
    :type vehicle_height: float

    :param vehicle_width:
        Vehicle width [m].
    :type vehicle_width: float

    :return:
        Raw frontal area of the vehicle [m2].
    :rtype: float
    """
    return vehicle_height * vehicle_width


@sh.add_function(dsp, outputs=['raw_frontal_area'], weight=5)
def calculate_raw_frontal_area_v1(vehicle_mass, vehicle_category):
    """
    Calculates raw frontal area of the vehicle [m2].

    :param vehicle_mass:
        Vehicle mass [kg].
    :type vehicle_mass: float

    :param vehicle_category:
        Vehicle category (i.e., A, B, C, D, E, F, S, M, and J).
    :type vehicle_category: str

    :return:
        Raw frontal area of the vehicle [m2].
    :rtype: float
    """
    from asteval import Interpreter as Interp
    d = dfl.functions.calculate_raw_frontal_area_v1
    expr = d.formulas[vehicle_category.upper()]
    return Interp(dict(vehicle_mass=vehicle_mass)).eval(expr)


@sh.add_function(dsp, outputs=['frontal_area'])
def calculate_frontal_area(raw_frontal_area):
    """
    Calculates the vehicle frontal area [m2].

    :param raw_frontal_area:
        Raw frontal area of the vehicle [m2].
    :type raw_frontal_area: float

    :return:
        Frontal area of the vehicle [m2].
    :rtype: float
    """
    d = dfl.functions.calculate_frontal_area.projection_factor
    return raw_frontal_area * d


@sh.add_function(dsp, outputs=['aerodynamic_drag_coefficient'])
def calculate_aerodynamic_drag_coefficient(vehicle_category):
    """
    Calculates the aerodynamic drag coefficient [-].

    :param vehicle_category:
        Vehicle category (i.e., A, B, C, D, E, F, S, M, and J).
    :type vehicle_category: str

    :return:
        Aerodynamic drag coefficient [-].
    :rtype: float
    """
    d = dfl.functions.calculate_aerodynamic_drag_coefficient
    return d.cw[vehicle_category.upper()]


@sh.add_function(dsp, outputs=['aerodynamic_drag_coefficient'])
def calculate_aerodynamic_drag_coefficient_v1(vehicle_body):
    """
    Calculates the aerodynamic drag coefficient [-].

    :param vehicle_body:
        Vehicle body (i.e., cabriolet, sedan, hatchback, stationwagon,
        suv/crossover, mpv, coupé, bus, bestelwagen, pick-up).
    :type vehicle_body: str

    :return:
        Aerodynamic drag coefficient [-].
    :rtype: float
    """
    d = dfl.functions.calculate_aerodynamic_drag_coefficient_v1
    return d.cw[vehicle_body.lower()]


dsp.add_data('air_temperature', dfl.values.air_temperature)
dsp.add_data('atmospheric_pressure', dfl.values.atmospheric_pressure)


@sh.add_function(dsp, outputs=['air_density'])
def calculate_air_density(air_temperature, atmospheric_pressure):
    """
    Calculates the air density [kg/m3].

    :param air_temperature:
        Air temperature [°C].
    :type air_temperature: float

    :param atmospheric_pressure:
        Atmospheric pressure [kPa].
    :type atmospheric_pressure: float

    :return:
        Air density [kg/m3].
    :rtype: float
    """
    # http://www.thecartech.com/subjects/auto_eng/Road_loads.htm
    return 3.48 * atmospheric_pressure / (273.16 + air_temperature)


dsp.add_data('has_roof_box', dfl.values.has_roof_box)


@sh.add_function(dsp, outputs=['aerodynamic_drag_area'], weight=4)
def calculate_aerodynamic_drag_value(
        aerodynamic_drag_coefficient, frontal_area, has_roof_box):
    """
    Calculates the aerodynamic drag area [m2].

    :param aerodynamic_drag_coefficient:
        Aerodynamic drag coefficient [-].
    :type aerodynamic_drag_coefficient: float

    :param frontal_area:
        Frontal area of the vehicle [m2].
    :type frontal_area: float

    :param has_roof_box:
         Has the vehicle a roof box? [-].
    :type has_roof_box: bool

    :return:
        Aerodynamic drag area [m2].
    :rtype: float
    """

    c = aerodynamic_drag_coefficient * frontal_area

    if has_roof_box:
        c *= dfl.functions.calculate_f2.roof_box

    return c


@sh.add_function(dsp, outputs=['f2'])
def calculate_f2(air_density, aerodynamic_drag_area):
    """
    Calculates the f2 coefficient [N/(km/h)^2].

    :param air_density:
        Air density [kg/m3].
    :type air_density: float

    :param aerodynamic_drag_area:
        Aerodynamic drag area [m2].
    :type aerodynamic_drag_area: float

    :return:
        As used in the dyno and defined by respective guidelines [N/(km/h)^2].
    :rtype: float
    """
    return aerodynamic_drag_area * (0.5 * air_density / 3.6 ** 2)


dsp.add_data('tyre_state', dfl.values.tyre_state)
dsp.add_data('road_state', dfl.values.road_state)


@sh.add_function(dsp, outputs=['static_friction'])
def default_static_friction(tyre_state, road_state):
    """
    Returns the default static friction coefficient [-].

    :param tyre_state:
        Tyre state (i.e., new or worm).
    :type tyre_state: str

    :param road_state:
        Road state (i.e., dry, wet, rainfall, puddles, ice).
    :type road_state: str

    :return:
        Static friction coefficient [-].
    :rtype: float
    """
    coeff = dfl.functions.default_static_friction.coeff
    return coeff[tyre_state][road_state]


@sh.add_function(dsp, outputs=['n_wheel'])
def default_n_wheel(n_wheel_drive):
    """
    Returns the default total number of wheels [-].

    :param n_wheel_drive:
        Number of wheel drive [-].
    :type n_wheel_drive: int

    :return:
        Total number of wheels [-].
    :rtype: int
    """
    return max(n_wheel_drive, dfl.functions.default_n_wheel.n_wheel)


@sh.add_function(dsp, outputs=['wheel_drive_load_fraction'])
def calculate_wheel_drive_load_fraction(n_wheel_drive, n_wheel=4):
    """
    Calculate the repartition of the load on wheel drive axles [-].

    :param n_wheel_drive:
        Number of wheel drive [-].
    :type n_wheel_drive: int

    :param n_wheel:
        Total number of wheels [-].
    :type n_wheel: int

    :return:
        Repartition of the load on wheel drive axles [-].
    :rtype: float
    """
    return n_wheel_drive / n_wheel


def _compile_traction_acceleration_limits(
        static_friction, wheel_drive_load_fraction):
    deceleration = -9.81 * static_friction
    acceleration = -deceleration * wheel_drive_load_fraction

    def _func(angle_slopes):
        slope = np.cos(angle_slopes)
        return deceleration * slope, acceleration * slope

    return _func


@sh.add_function(
    dsp, outputs=['traction_deceleration_limit', 'traction_acceleration_limit']
)
def calculate_traction_acceleration_limits(
        static_friction, wheel_drive_load_fraction, angle_slopes):
    """
    Calculates the traction acceleration limits [m/s2].

    :param static_friction:
        Static friction coefficient [-].
    :type static_friction: float

    :param wheel_drive_load_fraction:
        Repartition of the load on wheel drive axles [-].
    :type wheel_drive_load_fraction: float

    :param angle_slopes:
        Angle slope vector [rad].
    :type angle_slopes: numpy.array

    :return:
        Traction acceleration limits (i.e., deceleration, acceleration) [m/s2].
    :rtype: tuple[numpy.array]
    """
    return _compile_traction_acceleration_limits(
        static_friction, wheel_drive_load_fraction
    )(angle_slopes)


dsp.add_data('tyre_class', dfl.values.tyre_class)


@sh.add_function(dsp, outputs=['rolling_resistance_coeff'])
def calculate_rolling_resistance_coeff(tyre_class, tyre_category):
    """
    Calculates the rolling resistance coefficient [-].

    :param tyre_class: 
        Tyre class (i.e., C1, C2, and C3).
    :type tyre_class: str

    :param tyre_category: 
        Tyre category (i.e., A, B, C, D, E, F, and G).
    :type tyre_category: str

    :return: 
        Rolling resistance coefficient [-].
    :rtype: float
    """
    coeff = dfl.functions.calculate_rolling_resistance_coeff.coeff
    return coeff[tyre_class.upper()][tyre_category.upper()]


@sh.add_function(dsp, outputs=['f0'], weight=5)
def calculate_f0(vehicle_mass, rolling_resistance_coeff):
    """
    Calculates rolling resistance [N].

    :param vehicle_mass:
        Vehicle mass [kg].
    :type vehicle_mass: float

    :param rolling_resistance_coeff:
        Rolling resistance coefficient [-].
    :type rolling_resistance_coeff: float

    :return:
        Rolling resistance force [N] when angle_slope == 0.
    :rtype: float
    """
    return vehicle_mass * 9.81 * rolling_resistance_coeff


@sh.add_function(dsp, outputs=['f1'], weight=5)
def calculate_f1(f2):
    """
    Calculates the f1 road load [N/(km/h)].

    :param f2:
        As used in the dyno and defined by respective guidelines [N/(km/h)^2].
    :type f2: float

    :return: 
        Defined by dyno procedure [N/(km/h)].
    :rtype: float
    """

    q, m = dfl.functions.calculate_f1.qm
    return m * f2 + q


dsp.add_data('angle_slope', dfl.values.angle_slope)
dsp.add_data('path_elevations', wildcard=True)
dsp.add_data(
    'minimum_elevation_distance', dfl.values.minimum_elevation_distance
)


@sh.add_function(dsp, inputs_kwargs=True, outputs=['slope_model'])
@sh.add_function(
    dsp, inputs_kwargs=True, inputs=['path_distances', 'path_elevations'],
    outputs=['slope_model']
)
def define_slope_model(distances, elevations, minimum_elevation_distance=30):
    """
    Returns the angle slope model [rad].

    :param distances:
        Cumulative distance vector [m].
    :type distances: numpy.array

    :param elevations:
        Elevation vector [m].
    :type elevations: numpy.array

    :param minimum_elevation_distance:
        Minimum distance between points for the elevation interpolation [m].
    :type minimum_elevation_distance: numpy.array

    :return:
        Angle slope model [rad].
    :rtype: function
    """
    from scipy.interpolate import InterpolatedUnivariateSpline as Spl
    if minimum_elevation_distance > dfl.EPS:
        i, dm = [0, len(distances) - 1], distances[
            0] + minimum_elevation_distance
        for j, d in enumerate(distances[1:]):
            if d > dm:
                dm = distances[j] + minimum_elevation_distance
                i.append(j)
        i = np.unique(i)
    else:
        i = np.append([0], np.where(np.diff(distances) > dfl.EPS)[0] + 1)
    func = Spl(distances[i], elevations[i]).derivative()
    return lambda x: np.arctan(func(x))


@sh.add_function(dsp, outputs=['slope_model'], weight=5)
def define_slope_model_v1(angle_slope):
    """
    Returns the angle slope model [rad].

    :param angle_slope:
        Angle slope [rad].
    :type angle_slope: float

    :return:
        Angle slope model [rad].
    :rtype: function
    """

    def slope_model(distances):
        return np.full(distances.shape, angle_slope, float)

    return slope_model


@sh.add_function(dsp, outputs=['angle_slopes'])
def calculate_angle_slopes(slope_model, distances):
    """
    Returns the angle slope vector [rad].

    :param slope_model:
        Angle slope model [rad].
    :type slope_model: function

    :param distances:
        Cumulative distance vector [m].
    :type distances: numpy.array

    :return:
        Angle slope vector [rad].
    :rtype: numpy.array
    """
    return slope_model(distances)


@sh.add_function(dsp, outputs=['rolling_resistance'])
def calculate_rolling_resistance(f0, angle_slopes):
    """
    Calculates rolling resistance [N].

    :param f0:
        Rolling resistance force [N] when angle_slope == 0.
    :type f0: float

    :param angle_slopes:
        Angle slope vector [rad].
    :type angle_slopes: numpy.array

    :return:
        Rolling resistance force [N].
    :rtype: numpy.array
    """
    return f0 * np.cos(angle_slopes)


@sh.add_function(dsp, outputs=['velocity_resistances'])
def calculate_velocity_resistances(f1, velocities):
    """
    Calculates forces function of velocity [N].

    :param f1:
        Defined by dyno procedure [N/(km/h)].
    :type f1: float

    :param velocities:
        Velocity vector [km/h].
    :type velocities: numpy.array | float

    :return:
        Forces function of velocity [N].
    :rtype: numpy.array | float
    """

    return f1 * velocities


@sh.add_function(dsp, outputs=['climbing_force'])
def calculate_climbing_force(vehicle_mass, angle_slopes):
    """
    Calculates the vehicle climbing resistance [N].

    :param vehicle_mass:
        Vehicle mass [kg].
    :type vehicle_mass: float

    :param angle_slopes:
        Angle slope vector [rad].
    :type angle_slopes: numpy.array

    :return:
        Vehicle climbing resistance [N].
    :rtype: numpy.array
    """
    return vehicle_mass * 9.81 * np.sin(angle_slopes)


@sh.add_function(dsp, outputs=['n_dyno_axes'])
def select_default_n_dyno_axes(cycle_type, n_wheel_drive):
    """
    Selects the default number of dyno axes[-].

    :param cycle_type:
        Cycle type (WLTP or NEDC).
    :type cycle_type: str

    :param n_wheel_drive:
        Number of wheel drive [-].
    :type n_wheel_drive: int

    :return:
        Number of dyno axes [-].
    :rtype: int
    """
    par = dfl.functions.select_default_n_dyno_axes

    try:
        return par.DYNO_AXES[cycle_type.upper()][n_wheel_drive]
    except KeyError:
        return n_wheel_drive // 2


@sh.add_function(dsp, outputs=['inertial_factor'])
def select_inertial_factor(n_dyno_axes):
    """
    Selects the inertia factor [%] according to the number of dyno axes.

    :param n_dyno_axes:
        Number of dyno axes [-].
    :type n_dyno_axes: int

    :return:
        Factor that considers the rotational inertia [%].
    :rtype: float
    """

    return 1.5 * n_dyno_axes


@sh.add_function(dsp, outputs=['rotational_inertia_forces'])
def calculate_rotational_inertia_forces(
        vehicle_mass, inertial_factor, accelerations):
    """
    Calculate rotational inertia forces [N].

    :param vehicle_mass:
        Vehicle mass [kg].
    :type vehicle_mass: float

    :param inertial_factor:
        Factor that considers the rotational inertia [%].
    :type inertial_factor: float

    :param accelerations:
        Acceleration vector [m/s2].
    :type accelerations: numpy.array | float

    :return:
        Rotational inertia forces [N].
    :rtype: numpy.array | float
    """

    return vehicle_mass * inertial_factor * accelerations / 100


@sh.add_function(dsp, outputs=['accelerations'])
def calculate_accelerations_v1(
        vehicle_mass, inertial_factor, motive_forces, climbing_force,
        aerodynamic_resistances, rolling_resistance, velocity_resistances):
    """
    Calculates the acceleration from motive forces [m/s2].

    :param vehicle_mass:
        Vehicle mass [kg].
    :type vehicle_mass: float

    :param inertial_factor:
        Factor that considers the rotational inertia [%].
    :type inertial_factor: float

    :param motive_forces:
        Motive forces [N].
    :type motive_forces: numpy.array | float

    :param climbing_force:
        Vehicle climbing resistance [N].
    :type climbing_force: float | numpy.array

    :param aerodynamic_resistances:
        Aerodynamic resistance vector [N].
    :type aerodynamic_resistances: numpy.array | float

    :param rolling_resistance:
        Rolling resistance force [N].
    :type rolling_resistance: float | numpy.array

    :param velocity_resistances:
        Forces function of velocity [N].
    :type velocity_resistances: numpy.array | float

    :return:
        Acceleration vector [m/s2].
    :rtype: numpy.array
    """
    acc = motive_forces - climbing_force - aerodynamic_resistances
    acc -= rolling_resistance + velocity_resistances
    acc /= vehicle_mass * (1 + inertial_factor / 100)
    return acc


# noinspection PyPep8Naming
@sh.add_function(dsp, outputs=['motive_forces'])
def calculate_motive_forces(
        vehicle_mass, accelerations, climbing_force, aerodynamic_resistances,
        rolling_resistance, velocity_resistances, rotational_inertia_forces):
    """
    Calculate motive forces [N].

    :param vehicle_mass:
        Vehicle mass [kg].
    :type vehicle_mass: float

    :param accelerations:
        Acceleration vector [m/s2].
    :type accelerations: numpy.array | float

    :param climbing_force:
        Vehicle climbing resistance [N].
    :type climbing_force: float | numpy.array

    :param rolling_resistance:
        Rolling resistance force [N].
    :type rolling_resistance: float | numpy.array

    :param aerodynamic_resistances:
        Aerodynamic resistance vector [N].
    :type aerodynamic_resistances: numpy.array | float

    :param velocity_resistances:
        Forces function of velocity [N].
    :type velocity_resistances: numpy.array | float

    :param rotational_inertia_forces:
        Rotational inertia forces [N].
    :type rotational_inertia_forces: numpy.array | float

    :return:
        Motive forces [N].
    :rtype: numpy.array | float
    """

    # namespace shortcuts
    Frr = rolling_resistance
    Faero = aerodynamic_resistances
    Fclimb = climbing_force
    Fvel = velocity_resistances
    Finertia = rotational_inertia_forces

    return vehicle_mass * accelerations + Fclimb + Frr + Faero + Fvel + Finertia


@sh.add_function(dsp, outputs=['motive_powers'])
def calculate_motive_powers(motive_forces, velocities):
    """
    Calculates motive power [kW].

    :param motive_forces:
        Motive forces [N].
    :type motive_forces: numpy.array | float

    :param velocities:
        Velocity vector [km/h].
    :type velocities: numpy.array | float

    :return:
        Motive power [kW].
    :rtype: numpy.array | float
    """

    return motive_forces * velocities / 3600


@sh.add_function(dsp, outputs=['motive_forces'])
def calculate_motive_forces_v1(motive_powers, velocities):
    """
    Calculate motive forces [N].

    :param motive_powers:
        Motive power [kW].
    :type motive_powers: numpy.array | float

    :param velocities:
        Velocity vector [km/h].
    :type velocities: numpy.array | float

    :return:
        Motive forces [N].
    :rtype: numpy.array | float
    """

    return motive_powers / velocities * 3600


dsp.add_data(
    'road_loads', description='Cycle road loads [N, N/(km/h), N/(km/h)^2].'
)
dsp.add_function('grouping', sh.bypass, ['f0', 'f1', 'f2'], ['road_loads'])
dsp.add_function('splitting', sh.bypass, ['road_loads'], ['f0', 'f1', 'f2'])
dsp.add_data('correct_f0', dfl.values.correct_f0)


@sh.add_function(dsp, outputs=['f0'])
def apply_f0_correction(f0_uncorrected, correct_f0):
    """
    Corrects the rolling resistance force [N] if a different preconditioning
    cycle was used for WLTP (WLTP precon) and NEDC (NEDC precon).

    :param f0_uncorrected:
        Uncorrected rolling resistance force [N] when angle_slope == 0.
    :type f0_uncorrected: float

    :param correct_f0:
        A different preconditioning cycle was used for WLTP and NEDC?
    :type correct_f0: bool

    :return:
        Rolling resistance force [N] when angle_slope == 0.
    :rtype: float
    """

    if correct_f0:
        return f0_uncorrected - 6.0
    return f0_uncorrected


dsp.add_data('_f0', None, sh.inf(2, 0))
dsp.add_data('_f1', None, sh.inf(2, 0))
dsp.add_data('_f2', None, sh.inf(2, 0))
dsp.add_data('_vehicle_mass', None, sh.inf(2, 0))

dsp.add_function(function=sh.bypass, inputs=['f0'], outputs=['_f0'])
dsp.add_function(function=sh.bypass, inputs=['f1'], outputs=['_f1'])
dsp.add_function(function=sh.bypass, inputs=['f2'], outputs=['_f2'])
dsp.add_function(
    function=sh.bypass, inputs=['vehicle_mass'], outputs=['_vehicle_mass']
)


@sh.add_function(
    dsp, inputs_kwargs=True, outputs=['f0', 'f1', 'f2', 'vehicle_mass']
)
def identify_vehicle_mass_and_road_loads(
        inertial_factor, velocities, accelerations, angle_slopes, wheel_powers,
        _f0=None, _f1=None, _f2=None, _vehicle_mass=None):
    """
    Identifies vehicle road loads and its mass [N, N/(km/h), N/(km/h)^2, kg].

    :param inertial_factor:
        Factor that considers the rotational inertia [%].
    :type inertial_factor: float

    :param velocities:
        Velocity vector [km/h].
    :type velocities: numpy.array

    :param accelerations:
        Acceleration vector [m/s2].
    :type accelerations: numpy.array

    :param angle_slopes:
        Angle slope vector [rad].
    :type angle_slopes: numpy.array

    :param wheel_powers:
        Power at the wheels [kW].
    :type wheel_powers: numpy.array

    :param _f0:
        User input for rolling resistance force [N] when angle_slope == 0.
    :type _f0: float

    :param _f1:
        User input for f1.
    :type _f1: float

    :param _f2:
        User input for f2.
    :type _f2: float

    :param _vehicle_mass:
        User input for vehicle mass [kg].
    :type _vehicle_mass: float

    :return:
        The vehicle road loads and its mass [N, N/(km/h), N/(km/h)^2, kg].
    :rtype: tuple[float]
    """
    b = wheel_powers > 0
    matrix = np.column_stack((
        calculate_rolling_resistance(1, angle_slopes[b]),  # f0
        calculate_velocity_resistances(1, velocities[b]),  # f1
        calculate_aerodynamic_resistances(1, velocities[b]),  # f2
        calculate_climbing_force(1, angle_slopes[b])  # vehicle_mass

    ))
    matrix[:, 3] += accelerations[b] + calculate_rotational_inertia_forces(
        1, inertial_factor, accelerations[b]
    )
    matrix = calculate_motive_powers(matrix, velocities[b, None])
    x = {'f0': _f0, 'f1': _f1, 'f2': _f2, 'vehicle_mass': _vehicle_mass}
    mask, sol, y = np.ones((4,), bool), {}, wheel_powers[b]
    for i, (k, v) in enumerate(sorted(x.items())):
        if v is not None:
            mask[i] = False
            matrix[:, i] *= v
            y -= matrix[:, i]
            sol[k] = v
    matrix, keys = matrix[:, mask], sorted(set(x) - set(sol))
    return tuple(map(float, sh.selector(sorted(x), sh.combine_dicts(sh.map_list(
        keys, *np.linalg.lstsq(matrix, y, rcond=None)[0]
    ), base=sol), output_type='list')))



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the alternator current.
"""
import numpy as np
import schedula as sh
import co2mpas.utils as co2_utl

dsp = sh.BlueDispatcher(
    name='Alternator current', description='Models the alternator current.'
)


@sh.add_function(dsp, outputs=['alternator_current_model'])
def define_alternator_current_model(alternator_charging_currents):
    """
    Defines an alternator current model that predicts alternator current [A].

    :param alternator_charging_currents:
        Mean charging currents of the alternator (for negative and positive
        power)[A].
    :type alternator_charging_currents: (float, float)

    :return:
        Alternator current model.
    :rtype: callable
    """
    return AlternatorCurrentModel(alternator_charging_currents)


# noinspection PyMissingOrEmptyDocstring,PyPep8Naming
class AlternatorCurrentModel:
    def __init__(self, alternator_charging_currents=(0, 0)):
        def default_model(X):
            gb_power, acc =  X.T[-2:]
            b = gb_power > 0 or (gb_power == 0 and acc >= 0)

            return np.where(b, *alternator_charging_currents)

        # noinspection PyProtectedMember
        from ....engine._thermal import _XGBRegressor
        self.model = default_model
        self.mask = None
        self.init_model = default_model
        self.init_mask = None
        self.base_model = _XGBRegressor

    def _fit_charging_currents(self, X, y):
        gb_power, acc = X[:, -2:].T
        b = (gb_power > 0) | ((gb_power == 0) & (acc >= 0))
        if b.all() or ~b.all():
            return (np.median(y),) * 2
        return np.median(y[b]), np.median(y[~b])

    def predict(self, X, init_time=0.0):
        X = np.asarray(X)
        times = X[:, 0]
        b = times < (times[0] + init_time)
        curr = np.zeros_like(times, dtype=float)
        curr[b] = self.init_model(X[b][:, self.init_mask])
        b = ~b
        curr[b] = self.model(X[b][:, self.model])
        return curr

    # noinspection PyShadowingNames,PyProtectedMember
    def fit(self, currents, on_engine, times, soc, statuses, *args,
            init_time=0.0):
        b = (statuses[1:] > 0) & on_engine[1:]
        i = co2_utl.argmax(times >= times[0] + init_time)
        from ....engine._thermal import _build_samples
        X, Y = _build_samples(currents, soc, statuses, *args)
        if b[i:].any():
            self.model, self.mask = self._fit_model(X[i:][b[i:]], Y[i:][b[i:]])
            self.alternator_charging_currents = self._fit_charging_currents(
                X[i:][b[i:]], Y[i:][b[i:]]
            )
        elif b[:i].any():
            self.model, self.mask = self._fit_model(X[b], Y[b])
            self.alternator_charging_currents = self._fit_charging_currents(
                X[b], Y[b]
            )
        else:
            self.alternator_charging_currents = 0, 0
            self.model = lambda *args, **kwargs: [0.0]
            self.mask = np.array((0,))
        self.mask += 1

        if b[:i].any():
            init_spl = (times[1:i + 1] - times[0])[:, None], X[:i]
            init_spl = np.concatenate(init_spl, axis=1)[b[:i]]
            a = self._fit_model(init_spl, Y[:i][b[:i]], (0,), (2,))
            self.init_model, self.init_mask = a
        else:
            self.init_model, self.init_mask = self.model, self.mask

        return self

    # noinspection PyProtectedMember
    def _fit_model(self, X, Y, in_mask=(), out_mask=()):
        from sklearn.pipeline import Pipeline
        from ....engine._thermal import _SelectFromModel
        # noinspection PyArgumentEqualDefault
        model = self.base_model(
            random_state=0,
            max_depth=2,
            objective='reg:squarederror',
            n_estimators=int(min(300.0, 0.25 * (len(X) - 1))) or 1
        )
        model = Pipeline([
            ('feature_selection', _SelectFromModel(
                model, threshold='0.8*median', in_mask=in_mask,
                out_mask=out_mask
            )),
            ('classification', model)
        ])
        model.fit(X, Y)
        model.steps[0][1].estimator_.cache_params()
        model.steps[0][1].estimator.cache_params()
        model.steps[1][1].cache_params()
        mask = np.where(model.steps[0][-1]._get_support_mask())[0]
        return model.steps[-1][-1].predict, mask

    def __call__(self, time, soc, status, *args):
        arr = np.array([(time, soc, status) + args], float)
        if status == 3:
            return min(0.0, self.init_model(arr[:, self.init_mask])[0])
        return min(0.0, self.model(arr[:, self.mask])[0])


@sh.add_function(dsp, outputs=['alternator_current_model'])
def calibrate_alternator_current_model(
        alternator_currents, on_engine, times, service_battery_state_of_charges,
        service_battery_charging_statuses, motive_powers, accelerations,
        service_battery_initialization_time):
    """
    Calibrates an alternator current model that predicts alternator current [A].

    :param alternator_currents:
        Alternator currents [A].
    :type alternator_currents: numpy.array

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param service_battery_state_of_charges:
        State of charge of the service battery [%].
    :type service_battery_state_of_charges: numpy.array

    :param service_battery_charging_statuses:
        Service battery charging statuses (0: Discharge, 1: Charging, 2: BERS,
        3: Initialization) [-].
    :type service_battery_charging_statuses: numpy.array

    :param motive_powers:
        Motive power [kW].
    :type motive_powers: numpy.array

    :param accelerations:
        Acceleration vector [m/s2].
    :type accelerations: numpy.array

    :param service_battery_initialization_time:
        Service battery initialization time delta [s].
    :type service_battery_initialization_time: float

    :return:
        Alternator current model.
    :rtype: callable
    """
    return AlternatorCurrentModel().fit(
        alternator_currents, on_engine, times, service_battery_state_of_charges,
        service_battery_charging_statuses, motive_powers, accelerations,
        init_time=service_battery_initialization_time
    )


@sh.add_function(dsp, outputs=['alternator_currents'])
def predict_alternator_currents(
        alternator_current_model, times, service_battery_state_of_charges,
        service_battery_charging_statuses, motive_powers, accelerations,
        service_battery_initialization_time):
    """
    Predict alternator currents [A].

    :param alternator_current_model:
        Alternator current model.
    :type alternator_current_model: callable

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param service_battery_state_of_charges:
        State of charge of the service battery [%].
    :type service_battery_state_of_charges: numpy.array

    :param service_battery_charging_statuses:
        Service battery charging statuses (0: Discharge, 1: Charging, 2: BERS,
        3: Initialization) [-].
    :type service_battery_charging_statuses: numpy.array

    :param motive_powers:
        Motive power [kW].
    :type motive_powers: numpy.array

    :param accelerations:
        Acceleration vector [m/s2].
    :type accelerations: numpy.array

    :param service_battery_initialization_time:
        Service battery initialization time delta [s].
    :type service_battery_initialization_time: float

    :return:
        DC/DC converter current model.
    :rtype: callable
    """
    return alternator_current_model.predict(np.column_stack((
        times, service_battery_state_of_charges,
        service_battery_charging_statuses, motive_powers, accelerations
    )), init_time=service_battery_initialization_time)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and constants to define the co2_params selector.
"""
import copy
import schedula as sh
from co2mpas.utils import mae
from ._core import define_sub_model
from ...physical.engine.fc import dsp as _fc

#: Model name.
name = 'co2_params'

#: Parameters that constitute the model.
models = [
    'co2_params_calibrated', 'calibration_status', 'initial_friction_params',
    'engine_idle_fuel_consumption', 'kco2_wltp_correction_factor'
]

#: Inputs required to run the model.
inputs = ['co2_emissions_model']

#: Relevant outputs of the model.
outputs = ['co2_emissions', 'calibration_status']

#: Targets to compare the outputs of the model.
targets = ['identified_co2_emissions', 'calibration_status']

#: Weights coefficients to compute the model score.
weights = sh.map_list(targets, 1, None)


# noinspection PyUnusedLocal
def metric_calibration_status(y_true, y_pred):
    """
    Metric for the `calibration_status`.

    :param y_true:
        Reference calibration_status.
    :type y_true: list

    :param y_pred:
        Predicted calibration_status.
    :type y_pred: list

    :return:
        Error.
    :rtype: list
    """
    return [v[0] for v in y_pred]


#: Metrics to compare outputs with targets.
metrics = sh.map_list(targets, mae, metric_calibration_status)

#: Upper score limits to raise the warnings.
up_limit = {'identified_co2_emissions': 0.5}

#: Prediction model.
# noinspection PyProtectedMember
dsp = sh.Blueprint(copy.deepcopy(_fc).add_data(
    'kco2_wltp_correction_factor'
), inputs, outputs, models)._set_cls(define_sub_model)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and model `dsp` to model the mechanic of the clutch.
"""
import numpy as np
import schedula as sh
from co2mpas.defaults import dfl
from .torque_converter import define_k_factor_curve

dsp = sh.BlueDispatcher(name='Clutch', description='Models the clutch.')


@sh.add_function(dsp, outputs=['clutch_window'])
def default_clutch_window():
    """
    Returns a default clutching time window [s] for a generic clutch.

    :return:
        Clutching time window [s].
    :rtype: tuple
    """
    return dfl.functions.default_clutch_window.clutch_window


# noinspection PyUnusedLocal
def _no_model(times, **kwargs):
    return np.zeros_like(times)


# noinspection PyPep8Naming
@sh.add_function(dsp, outputs=['clutch_speed_model'])
def calibrate_clutch_speed_model(
        clutch_phases, accelerations, clutch_tc_speeds_delta, velocities,
        gear_box_speeds_in, gears):
    """
    Calibrate clutch speed model.

    :param clutch_phases:
        When the clutch is active [-].
    :type clutch_phases: numpy.array

    :param accelerations:
        Acceleration vector [m/s2].
    :type accelerations: numpy.array

    :param clutch_tc_speeds_delta:
        Engine speed delta due to the clutch [RPM].
    :type clutch_tc_speeds_delta: numpy.array

    :param velocities:
        Velocity vector [km/h].
    :type velocities: numpy.array

    :param gear_box_speeds_in:
        Gear box speed vector [RPM].
    :type gear_box_speeds_in: numpy.array

    :param gears:
        Gear vector [-].
    :type gears: numpy.array

    :return:
        Clutch speed model.
    :rtype: callable
    """
    model = _no_model
    if clutch_phases.sum() > 10:
        from co2mpas.utils import mae
        from sklearn.pipeline import Pipeline
        # noinspection PyProtectedMember
        from ..engine._thermal import _SelectFromModel, _XGBRegressor
        X = np.column_stack(
            (accelerations, velocities, gear_box_speeds_in, gears)
        )[clutch_phases]
        y = clutch_tc_speeds_delta[clutch_phases]

        # noinspection PyArgumentEqualDefault
        mdl = _XGBRegressor(
            max_depth=2,
            n_estimators=int(min(300., 0.25 * (len(y) - 1))),
            random_state=0,
            objective='reg:squarederror'
        )
        mdl = Pipeline([
            ('feature_selection',
             _SelectFromModel(mdl, threshold='0.8*median')),
            ('classification', mdl)
        ])
        mdl.fit(X, y)
        mdl.steps[0][1].estimator_.cache_params()
        mdl.steps[0][1].estimator.cache_params()
        mdl.steps[1][1].cache_params()
        if mae(mdl.predict(X), y) < mae(0, y):
            keys = 'accelerations', 'velocities', 'gear_box_speeds_in', 'gears'

            # noinspection PyUnusedLocal,PyMissingOrEmptyDocstring
            def model(times, **kwargs):
                return mdl.predict(np.column_stack(sh.selector(keys, kwargs)))
    return model


dsp.add_function(
    function=define_k_factor_curve,
    inputs=['stand_still_torque_ratio', 'lockup_speed_ratio'],
    outputs=['k_factor_curve']
)


@sh.add_function(dsp, outputs=['k_factor_curve'], weight=2)
def default_clutch_k_factor_curve():
    """
    Returns a default k factor curve for a generic clutch.

    :return:
        k factor curve.
    :rtype: callable
    """
    from co2mpas.defaults import dfl
    par = dfl.functions.default_clutch_k_factor_curve
    a = par.STAND_STILL_TORQUE_RATIO, par.LOCKUP_SPEED_RATIO
    from .torque_converter import define_k_factor_curve
    return define_k_factor_curve(*a)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
r"""
Define CO2MPAS command line interface.

.. click:: co2mpas.cli:cli
   :prog: co2mpas
   :show-nested:

"""
import os
import click
import logging
import click_log
import schedula as sh
import os.path as osp
from co2mpas import dsp as _process
from co2mpas._version import __version__

try:
    from co2mpas.cli.sync import cli as _sync
except ImportError:
    _sync = None
try:  # TODO: to be changed to co2mpas_gui.
    from co2wui.cli import cli as _gui
except ImportError:
    _gui = None

log = logging.getLogger('co2mpas.cli')
CO2MPAS_HOME = os.environ.get('CO2MPAS_HOME', '.')

log_config = dict(format="%(asctime)-15s:%(levelname)5.5s:%(name)s:%(message)s")


class _Logger(logging.Logger):
    # noinspection PyMissingOrEmptyDocstring
    def setLevel(self, level):
        super(_Logger, self).setLevel(level)
        logging.basicConfig(level=level, **log_config)
        rlog = logging.getLogger()
        # because `basicConfig()` does not reconfig root-logger when re-invoked.
        rlog.level = level
        logging.captureWarnings(True)


logger = _Logger('cli')
click_log.basic_config(logger)


@click.group(
    'co2mpas', context_settings=dict(help_option_names=['-h', '--help'])
)
@click.version_option(__version__)
def cli():
    """
    CO2MPAS command line tool.
    """


@cli.command('template', short_help='Generates input/output template file.')
@click.argument(
    'output-file', default='template.xlsx', required=False,
    type=click.Path(writable=True)
)
@click.option(
    '-TT', '--template-type', type=click.Choice(['input', 'output']),
    help='Template file type.', default='input', show_default=True
)
@click_log.simple_verbosity_option(logger)
def template(**inputs):
    """
    Writes a CO2MPAS input/output template into OUTPUT_FILE.

    OUTPUT_FILE: File path `.xlsx`. [default: ./template.xlsx]
    """
    return _process(inputs, ['template', 'done'])


@cli.command('demo', short_help='Generates sample demo files.')
@click.argument(
    'output-folder', default='./inputs', required=False,
    type=click.Path(writable=True, file_okay=False)
)
@click_log.simple_verbosity_option(logger)
def demo(output_folder):
    """
    Writes a CO2MPAS demo files into OUTPUT_FOLDER.

    OUTPUT_FOLDER: Folder path. [default: ./inputs]
    """
    return _process({'output_folder': output_folder}, ['demo', 'done'])


@cli.command('conf', short_help='Generates CO2MPAS model-configuration file.')
@click.argument(
    'output-file', default='./conf.yaml', required=False,
    type=click.Path(writable=True)
)
@click.option(
    '-MC', '--model-conf', type=click.Path(exists=True),
    help='Model-configuration file path `.yaml`.'
)
@click_log.simple_verbosity_option(logger)
def conf(output_file, **kwargs):
    """
    Writes a CO2MPAS model-configuration file into OUTPUT_FILE.

    OUTPUT_FILE: File path `.yaml`. [default: ./conf.yaml]
    """
    inputs = {sh.START: kwargs, 'output_file': output_file}
    return _process(inputs, ['conf', 'done'])


@cli.command('plot', short_help='Plots the CO2MPAS model.')
@click.option(
    '-C', '--cache-folder', help='Folder to save temporary html files.',
    default='./cache_plot', type=click.Path(file_okay=False, writable=True),
    show_default=True
)
@click.option(
    '-H', '--host', help='Hostname to listen on.', default='127.0.0.1',
    type=str, show_default=True
)
@click.option(
    '-P', '--port', help='Port of the webserver.', default=5000, type=int,
    show_default=True
)
@click_log.simple_verbosity_option(logger)
def plot(cache_folder, host, port):
    """
    Plots the full CO2MPAS model into CACHE_FOLDER.
    """
    return _process(
        dict(plot_model=True, cache_folder=cache_folder, host=host, port=port),
        ['plot', 'done']
    )


@cli.command('run', short_help='Run CO2MPAS model.')
@click.argument('input-files', nargs=-1, type=click.Path(exists=True))
@click.option(
    '-O', '--output-folder', help='Output folder.', default='./outputs',
    type=click.Path(file_okay=False, writable=True), show_default=True
)
@click.option(
    '-EK', '--encryption-keys', help='Encryption keys for TA mode.',
    default=osp.join(CO2MPAS_HOME, 'DICE_KEYS/dice.co2mpas.keys'),
    type=click.Path(), show_default=True
)
@click.option(
    '-SK', '--sign-key', help='User signature key for TA mode.',
    default=osp.join(CO2MPAS_HOME, 'DICE_KEYS/sign.co2mpas.key'),
    type=click.Path(), show_default=True
)
@click.option(
    '-C', '--cache-folder', help='Folder to save temporary html files.',
    default='./cache_plot', type=click.Path(file_okay=False, writable=True),
    show_default=True
)
@click.option(
    '-H', '--host', help='Hostname to listen on.', default='127.0.0.1',
    type=str, show_default=True
)
@click.option(
    '-P', '--port', help='Port of the webserver.', default=5000, type=int,
    show_default=True
)
@click.option(
    '-OT', '--output-template', help='Template output.',
    type=click.Path(exists=True)
)
@click.option(
    '-MC', '--model-conf', type=click.Path(exists=True),
    help='Model-configuration file path `.yaml`.'
)
@click.option(
    '-OS', '--only-summary', is_flag=True,
    help='Do not save vehicle outputs, just the summary.'
)
@click.option(
    '-AS', '--augmented-summary', is_flag=True,
    help='More outputs to the summary.'
)
@click.option(
    '-HV', '--hard-validation', is_flag=True, help='Add extra data validations.'
)
@click.option(
    '-DM', '--declaration-mode', is_flag=True,
    help='Use only the declaration data.'
)
@click.option(
    '-ES', '--enable-selector', is_flag=True,
    help='Enable the selection of the best model to predict both H/L cycles.'
)
@click.option(
    '-TA', '--type-approval-mode', is_flag=True, help='Is launched for TA?'
)
@click.option(
    '-PL', '--plot-workflow', is_flag=True,
    help='Open workflow-plot in browser, after run finished.'
)
@click.option(
    '-KP', '--encryption-keys-passwords',
    help='Encryption keys passwords file for reading TA files.',
    default='./DICE_KEYS/secret.passwords', type=click.Path(),
    show_default=True
)
@click_log.simple_verbosity_option(logger)
def run(input_files, cache_folder, host, port, plot_workflow, **kwargs):
    """
    Run CO2MPAS for all files into INPUT_FILES.

    INPUT_FILES: List of input files and/or folders
                 (format: .xlsx, .dill, .co2mpas.ta, .co2mpas).
    """
    kwargs = {k: v for k, v in kwargs.items() if v is not None}
    inputs = dict(
        plot_workflow=plot_workflow, host=host, port=port, cmd_flags=kwargs,
        input_files=input_files, cache_folder=cache_folder, **{sh.START: kwargs}
    )
    os.makedirs(inputs.get('output_folder') or '.', exist_ok=True)
    return _process(inputs, ['plot', 'done', 'run'])


if _sync is not None:
    cli.add_command(_sync, 'syncing')

if _gui is not None:
    cli.add_command(_gui, 'gui')

if __name__ == '__main__':
    cli()



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the alternator status.
"""
import numpy as np
import schedula as sh
from co2mpas.defaults import dfl
import co2mpas.utils as co2_utl

dsp = sh.BlueDispatcher(
    name='Service Battery Status model',
    description='Models the Service Battery Charging Status.'
)


# noinspection PyPep8
@sh.add_function(
    dsp, outputs=['service_battery_electric_powers_supply_threshold']
)
def define_service_battery_electric_powers_supply_threshold(
        service_battery_capacity, service_battery_nominal_voltage):
    """
    Identifies the service battery electric powers supply threshold [kW] that
    define when the the service battery is not charging.

    :param service_battery_capacity:
        Service battery capacity [Ah].
    :type service_battery_capacity: float

    :param service_battery_nominal_voltage:
        Service battery nominal voltage [V].
    :type service_battery_nominal_voltage: float

    :return:
        Service battery not charging power threshold [kW].
    :rtype: float
    """
    d = dfl.functions.define_service_battery_electric_powers_supply_threshold
    c = max(-service_battery_capacity * 36 * d.min_soc, d.min_current)
    return c * service_battery_nominal_voltage / 1e3


dsp.add_data(
    'service_battery_start_window_width',
    dfl.values.service_battery_start_window_width
)


@sh.add_function(dsp, outputs=['service_battery_starts_windows'])
def identify_service_battery_starts_windows(
        times, engine_starts, alternator_electric_powers,
        service_battery_start_window_width,
        service_battery_electric_powers_supply_threshold):
    """
    Identifies the alternator starts windows [-].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param engine_starts:
        When the engine starts [-].
    :type engine_starts: numpy.array

    :param alternator_electric_powers:
        Alternator electric power [kW].
    :type alternator_electric_powers: numpy.array

    :param service_battery_start_window_width:
        Service battery start window width [s].
    :type service_battery_start_window_width: float

    :param service_battery_electric_powers_supply_threshold:
        Service battery not charging power threshold [kW].
    :type service_battery_electric_powers_supply_threshold: float

    :return:
        Service battery starts windows [-].
    :rtype: numpy.array
    """
    ts, dt = times[engine_starts], service_battery_start_window_width / 2
    starts_windows = np.zeros_like(times, dtype=bool)
    ind = np.searchsorted(times, np.column_stack((ts - dt, ts + dt + dfl.EPS)))
    b = service_battery_electric_powers_supply_threshold
    b = alternator_electric_powers >= b
    for i, j in ind:
        starts_windows[i:j] = b[i:j].any()
    return starts_windows


def _set_alt_init_status(times, initialization_time, statuses):
    if initialization_time > 0:
        statuses[:co2_utl.argmax(times > (times[0] + initialization_time))] = 3
    return statuses


def _mask_boolean_phases(b):
    s = np.zeros(len(b) + 2, dtype=bool)
    s[1:-1] = b
    mask = np.column_stack((s[1:], s[:-1])) & (s[:-1] != s[1:])[:, None]
    return np.where(mask)[0].reshape((-1, 2))


@sh.add_function(dsp, outputs=['service_battery_charging_statuses'])
def identify_service_battery_charging_statuses(
        times, alternator_electric_powers, dcdc_converter_electric_powers,
        motive_powers, on_engine,
        service_battery_electric_powers_supply_threshold,
        service_battery_starts_windows, service_battery_initialization_time):
    """
    Identifies service battery charging statuses: Discharge (0), Charging (1),
    BERS (2), and Initialization(3) [-].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param alternator_electric_powers:
        Alternator electric power [kW].
    :type alternator_electric_powers: numpy.array

    :param dcdc_converter_electric_powers:
        DC/DC converter electric power [kW].
    :type dcdc_converter_electric_powers: numpy.array

    :param motive_powers:
        Motive power [kW].
    :type motive_powers: numpy.array

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :param service_battery_electric_powers_supply_threshold:
        Service battery not charging power threshold [kW].
    :type service_battery_electric_powers_supply_threshold: float

    :param service_battery_starts_windows:
        Service battery starts windows [-].
    :type service_battery_starts_windows: numpy.array

    :param service_battery_initialization_time:
        Service battery initialization time delta [s].
    :type service_battery_initialization_time: float

    :return:
        Service battery charging statuses (0: Discharge, 1: Charging, 2: BERS,
        3: Initialization) [-].
    :rtype: numpy.array
    """
    threshold = service_battery_electric_powers_supply_threshold
    b = (alternator_electric_powers < threshold) & on_engine
    status = b.astype(int, copy=True)
    status[b & (motive_powers < 0)] = 2

    off = ~on_engine | service_battery_starts_windows
    mask = _mask_boolean_phases(status != 1)
    for i, j in mask:
        # noinspection PyUnresolvedReferences
        if ((status[i:j] == 2) | off[i:j]).all():
            status[i:j] = 1

    status[dcdc_converter_electric_powers < threshold] = 1
    _set_alt_init_status(times, service_battery_initialization_time, status)

    return status


# noinspection PyPep8
@sh.add_function(dsp, outputs=['service_battery_initialization_time'])
def identify_service_battery_initialization_time(
        alternator_electric_powers, motive_powers,
        accelerations, service_battery_state_of_charges,
        service_battery_charging_statuses, times,
        service_battery_electric_powers_supply_threshold):
    """
    Identifies the alternator initialization time delta [s].

    :param alternator_electric_powers:
        Alternator electric power [kW].
    :type alternator_electric_powers: numpy.array

    :param motive_powers:
        Motive power [kW].
    :type motive_powers: numpy.array

    :param accelerations:
        Vehicle acceleration [m/s2].
    :type accelerations: numpy.array

    :param service_battery_state_of_charges:
        State of charge of the service battery [%].
    :type service_battery_state_of_charges: numpy.array

    :param service_battery_charging_statuses:
        Service battery charging statuses (0: Discharge, 1: Charging, 2: BERS,
        3: Initialization) [-].
    :type service_battery_charging_statuses: numpy.array

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param service_battery_electric_powers_supply_threshold:
        Service battery not charging power threshold [kW].
    :type service_battery_electric_powers_supply_threshold: float

    :return:
        Service battery initialization time delta [s].
    :rtype: float
    """
    bats, p = service_battery_charging_statuses, motive_powers
    i = co2_utl.argmax(bats != 0)
    if bats[0] == 1 or (i and ((bats[:i] == 0) & (p[:i] == 0)).all()):
        s = service_battery_electric_powers_supply_threshold
        s = alternator_electric_powers < s
        n, i = len(times), int(co2_utl.argmax((s[:-1] != s[1:]) & s[:-1]))
        i = min(n - 1, i)

        # noinspection PyProtectedMember
        from ....engine._thermal import _build_samples, _XGBRegressor

        x, y = _build_samples(
            alternator_electric_powers, service_battery_state_of_charges, bats,
            p, accelerations
        )

        j = min(i, int(n / 2))
        # noinspection PyArgumentEqualDefault
        model = _XGBRegressor(
            random_state=0,
            max_depth=2,
            n_estimators=int(min(100.0, 0.25 * (n - j))) or 1,
            objective='reg:squarederror'
        ).fit(x[j:], y[j:])
        err = np.abs(y - model.predict(x))
        sets = np.array(co2_utl.get_inliers(err)[0], dtype=int)[:i]
        if (i and sum(sets) / i < 0.5) or i > j:
            from sklearn.tree import DecisionTreeClassifier
            reg = DecisionTreeClassifier(max_depth=1, random_state=0)
            reg.fit(times[1:i + 1, None], sets)
            s, r = reg.tree_.children_left[0], reg.tree_.children_right[0]
            s, r = np.argmax(reg.tree_.value[s]), np.argmax(reg.tree_.value[r])
            if s == 0 and r == 1:
                return reg.tree_.threshold[0] - times[0]
            elif r == 0 and not i > j:
                return times[i] - times[0]

    elif bats[0] == 3:
        i = co2_utl.argmax(bats != 3)
        return times[i] - times[0]
    return 0.0


@sh.add_function(dsp, outputs=[
    'service_battery_charging_statuses', 'service_battery_initialization_time'
], weight=1)
def identify_service_battery_charging_statuses_and_initialization_time(
        times, accelerations, on_engine, alternator_electric_powers,
        dcdc_converter_electric_powers, motive_powers,
        service_battery_electric_powers_supply_threshold,
        service_battery_starts_windows, service_battery_state_of_charges):
    """
    Identifies the service battery charging statuses [-] and its initialization
    time delta [s].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param alternator_electric_powers:
        Alternator electric power [kW].
    :type alternator_electric_powers: numpy.array

    :param dcdc_converter_electric_powers:
        DC/DC converter electric power [kW].
    :type dcdc_converter_electric_powers: numpy.array

    :param motive_powers:
        Motive power [kW].
    :type motive_powers: numpy.array

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :param service_battery_electric_powers_supply_threshold:
        Service battery not charging power threshold [kW].
    :type service_battery_electric_powers_supply_threshold: float

    :param service_battery_starts_windows:
        Service battery starts windows [-].
    :type service_battery_starts_windows: numpy.array

    :param accelerations:
        Acceleration [m/s2].
    :type accelerations: numpy.array

    :param service_battery_state_of_charges:
        State of charge of the service battery [%].
    :type service_battery_state_of_charges: numpy.array

    :return:
       Service battery charging statuses (0: Discharge, 1: Charging, 2: BERS,
        3: Initialization) [-] and its initialization time delta [s].
    :rtype: numpy.array, float
    """
    statuses = identify_service_battery_charging_statuses(
        times, alternator_electric_powers, dcdc_converter_electric_powers,
        motive_powers, on_engine,
        service_battery_electric_powers_supply_threshold,
        service_battery_starts_windows, 0
    )
    initialization_time = identify_service_battery_initialization_time(
        alternator_electric_powers, motive_powers, accelerations,
        service_battery_state_of_charges, statuses, times,
        service_battery_electric_powers_supply_threshold
    )
    _set_alt_init_status(times, initialization_time, statuses)
    return statuses, initialization_time


def _identify_balance_soc(times, state_of_charges):
    import lmfit
    parameters = lmfit.Parameters()
    parameters.add('B', value=np.median(state_of_charges), min=0.0, max=100.0)
    parameters.add('A', value=0)
    parameters.add('X0', value=1.0, min=0.0)
    x = (times - times[0]) / (times[-1] - times[0])
    n = len(x)

    # noinspection PyMissingOrEmptyDocstring
    def func(params):
        p = params.valuesdict()
        y = np.tile(p['B'], n)
        b = x < p['X0']
        y[b] += p['A'] * (x[b] - p['X0']) ** 2
        return y

    # noinspection PyMissingOrEmptyDocstring
    def error(params):
        return co2_utl.mae(state_of_charges, func(params))

    # noinspection PyProtectedMember
    from ....engine.fc import _calibrate_model_params
    return _calibrate_model_params(error, parameters)[0].valuesdict()['B']


# noinspection PyMissingOrEmptyDocstring,PyPep8Naming
class BatteryStatusModel:
    def __init__(self, bers_pred=None, charge_pred=None, min_soc=0.0,
                 max_soc=100.0):
        self.charge = charge_pred or (lambda X: np.zeros(len(X), dtype=bool))
        self.bers = bers_pred
        self.max = max_soc
        self.min = min_soc

    def __call__(self, *args, **kwargs):
        return self.predict(*args, **kwargs)

    def _fit_bers(self, charging_statuses, motive_powers):
        b = charging_statuses == 2
        threshold = 0.0
        if b.any():
            from sklearn.tree import DecisionTreeClassifier
            q = dfl.functions.BatteryStatusModel.min_percentile_bers
            m = DecisionTreeClassifier(random_state=0, max_depth=2)
            c = charging_statuses != 1
            # noinspection PyUnresolvedReferences
            m.fit(motive_powers[c, None], b[c])

            X = motive_powers[b, None]
            if (np.sum(m.predict(X)) / X.shape[0] * 100) >= q:
                self.bers = m.predict  # shortcut name
                return self.bers

            # noinspection PyUnresolvedReferences
            if not b.all():
                gb_p_s = motive_powers[_mask_boolean_phases(b)[:, 0]]

                threshold = min(threshold, np.percentile(gb_p_s, q))

        self.bers = lambda x: np.asarray(x) < threshold
        return self.bers

    # noinspection PyShadowingNames
    def _fit_charge(self, charging_statuses, state_of_charges, times,
                    is_hybrid):
        b = charging_statuses[1:] == 1
        self.max, self.min = 100.0, 0.0
        if b.all():
            self.min = max(state_of_charges)
        elif b.any():
            from sklearn.tree import DecisionTreeClassifier
            charge = DecisionTreeClassifier(random_state=0, max_depth=3)
            X = np.column_stack(
                (charging_statuses[:-1], state_of_charges[1:])
            )
            charge.fit(X, b)
            soc = state_of_charges[charging_statuses == 1]
            X = np.column_stack((
                np.ones(100), np.linspace(soc.min(), soc.max(), 100)
            ))
            s = np.where(charge.predict(X))[0]
            if is_hybrid:
                self.charge = lambda x: [x[0][0] == 1]
            else:
                self.charge = charge.predict
            if s.shape[0]:
                self.min, self.max = X[s[0], 1], X[s[-1], 1]
        self._fit_boundaries(charging_statuses, state_of_charges, times)

    def _fit_boundaries(self, charging_statuses, state_of_charges, times):
        n, b = len(charging_statuses), charging_statuses == 1
        mask = _mask_boolean_phases(b)
        _max, _min, balance = [], [], ()
        from scipy.stats import linregress
        min_dt = dfl.functions.BatteryStatusModel.min_delta_time_boundaries
        for i, j in mask:
            t = times[i:j]
            if t[-1] - t[0] <= min_dt:
                continue
            soc = state_of_charges[i:j]
            m, q = linregress(t, soc)[:2]
            if m >= 0:
                if i > 0:
                    _min.append(soc.min())
                if j < n:
                    _max.append(soc.max())

        min_delta_soc = dfl.functions.BatteryStatusModel.min_delta_soc
        if _min:
            self.min = _min = max(self.min, min(100.0, min(_min)))

            _max = [m for m in _max if m >= _min]
            if _max:
                self.max = min(100.0, min(max(_max), _min + min_delta_soc))
            else:
                self.max = min(100.0, _min + min_delta_soc)
        elif _max:
            self.max = _max = min(self.max, max(0.0, max(_max)))
            self.min = _max - min_delta_soc
        elif b.any() and not b.all():
            balance = _identify_balance_soc(times, state_of_charges)
            # noinspection PyTypeChecker
            std = np.sqrt(np.mean((balance - state_of_charges) ** 2)) * 2
            std = min(min_delta_soc, std)
            self.max = min(balance + std, 100.0)
            self.min = max(balance - std, 0.0)

    def fit(self, is_hybrid, times, charging_statuses, state_of_charges,
            motive_powers):

        i = co2_utl.argmax(charging_statuses != 3)

        status, soc = charging_statuses[i:], state_of_charges[i:]

        self._fit_bers(status, motive_powers[i:])
        self._fit_charge(status, soc, times[i:], is_hybrid)

        return self

    def predict(self, has_energy_rec, init_time, time, prev, soc, power):
        status = 0

        if soc < 100:
            func = self.charge
            if time < init_time:
                status = 3

            elif soc < self.min or (soc <= self.max and func([[prev, soc]])[0]):
                status = 1

            elif has_energy_rec and self.bers([(power,)])[0]:
                status = 2

        return status


@sh.add_function(dsp, outputs=['service_battery_status_model'], weight=10)
def calibrate_service_battery_status_model(
        is_hybrid, times, service_battery_charging_statuses,
        service_battery_state_of_charges, motive_powers):
    """
    Calibrates the service battery charging status model.

    :param is_hybrid:
        Is the vehicle hybrid?
    :type is_hybrid: bool

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param service_battery_charging_statuses:
        Service battery charging statuses (0: Discharge, 1: Charging, 2: BERS,
        3: Initialization) [-].
    :type service_battery_charging_statuses: numpy.array

    :param service_battery_state_of_charges:
        State of charge of the service battery [%].
    :type service_battery_state_of_charges: numpy.array

    :param motive_powers:
        Motive power [kW].
    :type motive_powers: numpy.array

    :return:
        A function that predicts the service battery charging status.
    :rtype: callable
    """
    return BatteryStatusModel().fit(
        is_hybrid, times, service_battery_charging_statuses,
        service_battery_state_of_charges, motive_powers
    )


@sh.add_function(dsp, outputs=['service_battery_status_model'])
def define_service_battery_status_model(
        service_battery_state_of_charge_balance,
        service_battery_state_of_charge_balance_window):
    """
    Defines the service battery charging status model.

    :param service_battery_state_of_charge_balance:
        Service battery state of charge balance [%].
    :type service_battery_state_of_charge_balance: float

    :param service_battery_state_of_charge_balance_window:
        Service battery state of charge balance [%].
    :type service_battery_state_of_charge_balance_window: float

    :return:
        A function that predicts the service battery charging status.
    :rtype: callable
    """
    m = service_battery_state_of_charge_balance
    w = service_battery_state_of_charge_balance_window / 2
    return BatteryStatusModel(
        charge_pred=lambda x: [x[0][0] == 1],
        bers_pred=lambda x: [x[0][0] < 0],
        min_soc=m - w, max_soc=m + w
    )


# noinspection PyPep8Naming
@sh.add_function(dsp, outputs=[
    'service_battery_state_of_charge_balance',
    'service_battery_state_of_charge_balance_window'
])
def identify_service_battery_state_of_charge_balance_and_window(
        service_battery_status_model):
    """
    Identify the service battery state of charge balance and window [%].

    :param service_battery_status_model:
        A function that predicts the service battery charging status.
    :type service_battery_status_model: BatteryStatusModel

    :return:
        Service battery state of charge balance and window [%].
    :rtype: float, float
    """
    model = service_battery_status_model
    state_of_charge_balance_window = model.max - model.min
    state_of_charge_balance = model.min + state_of_charge_balance_window / 2
    return state_of_charge_balance, state_of_charge_balance_window



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions to read inputs from excel.
"""
import io
import math
import regex
import logging
import numpy as np
import os.path as osp
import schedula as sh
from collections.abc import Iterable
from xlref.parser import Ref, _re_xl_ref_parser, FILTERS

log = logging.getLogger(__name__)

_base_params = r"""
    ^((?P<scope>base)[.\s]+)?
    ((?P<usage>(target|input|output|data))s?[.\s]+)?
    ((?P<stage>(calibration|prediction))s?[.\s]+)?
    ((?P<cycle>(WLTP|NEDC|ALL)([-_][HLM])?)[.\s]+)?
    (?P<param>[^\s.]*)\s*$
    |
    ^((?P<scope>base)[.\s]+)?
    ((?P<usage>(target|input|output|data))s?[.\s]+)?
    ((?P<stage>(calibration|prediction))s?[.\s]+)?
    ((?P<param>[^\s.]*))?
    ((.|\s+)(?P<cycle>(WLTP|NEDC|ALL)([-_][HLM])?))?\s*$
"""

_flag_params = r"""
^(?P<scope>flag)[.\s]+(?P<flag>(input_version|vehicle_family_id))([.\s]+ALL)?\s*$
"""

_dice_params = r"""
    ^(?P<scope>dice)[.\s]+(?P<dice>.+)[.\s]+ALL\s*$
    |
    ^(?P<scope>dice)[.\s]+(?P<dice>.+)\s*$    
"""

_meta_params = r"""
    ^(?P<scope>meta)[.\s]+((?P<meta>.+)[.\s]+)?((?P<param>[^\s.]+))
    ((.|\s+)(?P<cycle>(WLTP|NEDC|ALL)([-_][HLM])?))?\s*$
"""

_plan_params = r"""
    ^(?P<scope>plan)[.\s]+(
     (?P<index>(id|base|run_base))\s*$
     |
""" + _flag_params.replace('<scope>', '<v_scope>').replace('^(', '(') + r"""
     |
""" + _dice_params.replace('<scope>', '<v_scope>').replace('^(', '(') + r"""
     |
""" + _meta_params.replace('<scope>', '<v_scope>').replace('^(', '(') + r"""
     |
""" + _base_params.replace('<scope>', '<v_scope>').replace('^(', '(') + r"""
     )
"""

_re_params_name = regex.compile(
    r"""
        ^(?P<param>((plan|base|flag|dice)|
                    (target|input|output|data|meta)|
                    ((calibration|prediction)s?)|
                    ((WLTP|NEDC|ALL)([-_][HLM])?)))\s*$
        |
    """ + _flag_params + r"""
        |
    """ + _dice_params + r"""
        |
    """ + _meta_params + r"""
        |
    """ + _plan_params + r"""
        |
    """ + _base_params, regex.IGNORECASE | regex.X | regex.DOTALL)

_base_sheet = r"""
    ^((?P<scope>base)([.\s]+)?)?
    ((?P<usage>(target|input|output|data))s?([.\s]+)?)?
    ((?P<stage>(calibration|prediction))s?([.\s]+)?)?
    ((?P<cycle>(WLTP|NEDC|ALL)([-_][HLM])?)([.\s]+)?)?
    (?P<type>(pa|ts|pl|mt))?\s*$
"""

_flag_sheet = r"""^(?P<scope>flag)([.\s]+(?P<type>(pa|ts|pl|mt)))?\s*$"""

_dice_sheet = r"""^(?P<scope>dice)([.\s]+(?P<type>(pa|ts|pl|mt)))?\s*$"""

_meta_sheet = r"""
    ^(?P<scope>meta)([.\s]+(?P<meta>.+))?[.\s]+(?P<type>(pa|ts|pl|mt))\s*$
"""

_plan_sheet = r"""
    ^(?P<scope>plan)([.\s]+(
""" + _flag_sheet.replace('<scope>', '<v_scope>').replace('^(', '(') + r"""
     |
""" + _dice_sheet.replace('<scope>', '<v_scope>').replace('^(', '(') + r"""
     |
""" + _meta_sheet.replace('<scope>', '<v_scope>').replace('^(', '(') + r"""
     |
""" + _base_sheet.replace('<scope>', '<v_scope>').replace('^(', '(') + r"""
     ))?\s*$
"""

_re_input_sheet_name = regex.compile(r'|'.join(
    (_flag_sheet, _dice_sheet, _meta_sheet, _plan_sheet, _base_sheet)
), regex.IGNORECASE | regex.X | regex.DOTALL)

_re_space_dot = regex.compile(r'(\s*\.\s*|\s+)')

_xl_ref = {
    'pa': '#%s!B2:C_[{"fun": "dict", "key": "ref", "value": "ref"}]',
    'ts': '#%s!A2(R):._:RD["T", "dict"]',
    'pl': '#%s!A1(R):._:R[{"fun": "recursive", "dtype": "object"}]',
    'mt': '#%s!B1:._:R[{"fun": "recursive", "dtype": "object"}, "matrix"]'
}


def _matrix(parent, x):
    from pandas import isnull
    b, f = np.append([False], ~isnull(x[0][1:])), '{}.{}'.format
    keys = [
        None if isnull(k) else f(k.strip(' '), c.strip(' '))
        for k in x[1:, 0] for c in x[0][b]]

    return dict(zip(keys, x[1:, b].ravel()))


def _vector(parent, x):
    from pandas import notnull
    x = x.ravel()
    return x if notnull(x).any() else []


def _idict(parent, x):
    return dict(enumerate(x, 1))


def _empty(parent, x):
    return x if len(x) else None


FILTERS['empty'] = _empty
FILTERS['idict'] = _idict
FILTERS['vector'] = _vector
FILTERS['matrix'] = _matrix


class Rererence(Ref):
    _re = regex.compile(
        _re_xl_ref_parser.pattern.replace('?P<filters>', r'?P<filters>:?\s*'),
        _re_xl_ref_parser.flags
    )

    def _match(self, ref):
        d = super(Rererence, self)._match(ref)
        if (d.get('filters') or '').startswith(':'):
            d['filters'] = None
            if d['nd_col'] == d['nd_row'] == '.' and d['nd_mov']:
                d['range_exp'] = d['nd_mov'] + (d['range_exp'] or '')
                d['nd_mov'] = None
        return d


# noinspection PyShadowingBuiltins,PyUnusedLocal
def _get_sheet_type(
        type=None, usage=None, cycle=None, scope='base', **kw):
    if type:
        pass
    elif scope == 'plan':
        type = 'pl'
    elif scope in ('flag', 'dice') or not cycle:
        type = 'pa'
    else:
        type = 'ts'
    return type


def _check_none(v):
    if v is None:
        return True
    elif isinstance(v, Iterable) and not isinstance(v, str) \
            and len(v) <= 1:
        # noinspection PyTypeChecker
        return _check_none(next(iter(v))) if len(v) == 1 else True
    return False


def _isempty(val):
    return isinstance(val, float) and math.isnan(val) or _check_none(val)


# noinspection PyUnusedLocal
def _get_cycle(cycle=None, usage=None, **kw):
    if cycle is None or cycle == 'all':
        cycle = 'nedc_h', 'nedc_l', 'wltp_h', 'wltp_l', 'wltp_m'
    elif cycle == 'wltp':
        cycle = 'wltp_h', 'wltp_l', 'wltp_m'
    elif cycle == 'nedc':
        cycle = 'nedc_h', 'nedc_l'
    elif cycle in ('all-h', 'all_h'):
        cycle = 'nedc_h', 'wltp_h'
    elif cycle in ('all-l', 'all_l'):
        cycle = 'nedc_l', 'wltp_l'
    elif cycle in ('all-m', 'all_m'):
        cycle = 'wltp_m',
    elif isinstance(cycle, str):
        cycle = cycle.replace('-', '_')

    return cycle


# noinspection PyUnusedLocal
def _get_default_stage(stage=None, cycle=None, usage=None, **kw):
    if stage is None:
        if 'nedc' in cycle or usage == 'target':
            stage = 'prediction'
        else:
            stage = 'calibration'

    return stage.replace(' ', '')


def _parse_key(scope='base', usage='input', **match):
    if scope == 'flag':
        if match['flag'] == 'vehicle_family_id':
            scope = 'dice'
        yield scope, match['flag']
    elif scope == 'dice':
        yield scope, match['dice']
        if match['dice'] == 'input_type':
            yield from _parse_key(cycle='all', param='input_type')
    elif scope == 'meta':
        meta = _re_space_dot.sub(match.get('meta', ''), '.').replace('-', '_')
        param = match['param']
        if match.get('cycle', 'ALL') != 'ALL':
            param = match['cycle'] + param
        yield scope, meta, param
    elif scope == 'plan':
        if 'param' in match:
            m = _re_params_name.match('.'.join((scope, match['param'])))
            if m:
                m = {i: j for i, j in m.groupdict().items() if j}
                if 'index' in m:
                    match = m

        if 'index' in match:
            yield scope, match['index']
        else:
            for k in _parse_key(match.get('v_scope', 'base'), usage, **match):
                yield scope, '.'.join(k)
    elif scope == 'base':
        i = match['param']

        if i.lower() == 'version':
            yield 'flag', 'input_version'
        else:
            m = match.copy()
            for c in sh.stlp(_get_cycle(usage=usage, **match)):
                m['cycle'] = c
                stage = _get_default_stage(usage=usage, **m)
                yield scope, usage, stage, c, i


def _parse_values(data, default=None, where=''):
    default = default or {}
    for k, v in data.items():
        k = k.strip(' ')
        match = _re_params_name.match(k) if k is not None else None
        if not match and default.get('scope') == 'meta':
            # noinspection PyTypeChecker
            match = _re_params_name.match(
                '.'.join(filter(bool, ('meta', default.get('meta'), k)))
            )
        if not match:
            log.warning("Parameter '%s' %s cannot be parsed!", k, where)
            continue
        elif _isempty(v):
            continue
        match = {i: j.lower() for i, j in match.groupdict().items() if j}

        for key in _parse_key(**sh.combine_dicts(default, match)):
            yield key, v


def _parse_sheet(match, parent, sheet_name, res=None):
    if res is None:
        res = {}

    import pandas as pd
    sh_type = _get_sheet_type(**match)
    data = Rererence(_xl_ref[sh_type] % sheet_name, parent, parent.cache).values
    if sh_type == 'pl':
        try:
            data = pd.DataFrame(data[1:], columns=data[0])
        except IndexError:
            return None
        if 'id' not in data:
            data['id'] = data.index + 1
        else:
            data['id'] = data['id'].apply(
                lambda x: x.strip() if isinstance(x, str) else x
            )

        data.set_index(['id'], inplace=True)
        data.dropna(how='all', inplace=True)
        data.dropna(axis=1, how='all', inplace=True)
    elif sh_type == 'ts':
        data = pd.DataFrame(data)
        data.dropna(how='all', inplace=True)
        data.dropna(axis=1, how='all', inplace=True)
        # noinspection PyProtectedMember
        mask = data.count(0) == len(data._get_axis(0))
        # noinspection PyUnresolvedReferences
        drop = [k for k, v in mask.items() if not v]
        if drop:
            msg = 'Columns {} in {} sheet contains nan.\n ' \
                  'Please correct the inputs!'
            raise ValueError(msg.format(drop, sheet_name))
        data = data.to_dict('list')
    else:
        import numpy as np
        data = {
            k: v for k, v in data.items()
            if k and not pd.isnull(np.ravel(v)).any()
        }

    for k, v in _parse_values(data, match, "in sheet '%s'" % sheet_name):
        sh.get_nested_dicts(res, *k[:-1])[k[-1]] = v
    return res


def _add_times_base(data, scope='base', usage='input', **match):
    if scope != 'base':
        return
    sh_type = _get_sheet_type(scope=scope, usage=usage, **match)
    n = (scope, 'target')
    if sh_type == 'ts' and sh.are_in_nested_dicts(data, *n):
        t = sh.get_nested_dicts(data, *n)
        for k, v in sh.stack_nested_keys(t, key=n, depth=2):
            if 'times' not in v:
                n = list(k + ('times',))
                n[1] = usage
                if sh.are_in_nested_dicts(data, *n):
                    v['times'] = sh.get_nested_dicts(data, *n)
                else:
                    for i, j in sh.stack_nested_keys(data, depth=4):
                        if 'times' in j:
                            v['times'] = j['times']
                            break


def _add_index_plan(plan, file_path):
    if 'base' not in plan:
        plan['base'] = file_path
    else:
        d = osp.dirname(file_path)
        plan['base'].fillna(osp.basename(file_path), inplace=True)
        plan['base'] = plan['base'].apply(
            lambda x: osp.isabs(x) and x or osp.join(d, x)
        )

    plan['base'] = plan['base'].apply(osp.normpath)

    if 'run_base' not in plan:
        plan['run_base'] = True
    else:
        plan['run_base'].fillna(True)

    plan['id'] = plan.index
    return plan


def _finalize_plan(res, plans, file_path):
    import pandas as pd
    if not plans:
        plans = (pd.DataFrame(),)

    for k, v in sh.stack_nested_keys(res.get('plan', {}), depth=4):
        n = '.'.join(k)
        m = '.'.join(k[:-1])
        for p in plans:
            if any(c.startswith(m) for c in p.columns):
                if n in p:
                    p[n].fillna(value=v, inplace=True)
                else:
                    p[n] = v

    plan = pd.concat(plans, axis=1, copy=False, verify_integrity=True)
    # noinspection PyTypeChecker
    return _add_index_plan(plan, file_path)


def parse_excel_file(input_file_name, input_file):
    """
    Reads cycle's data and simulation plans.

    :param input_file_name:
        Input file name.
    :type input_file_name: str

    :param input_file:
        Input file.
    :type input_file: io.BytesIO

    :return:
        Raw input data.
    :rtype: dict
    """
    import xlref
    import warnings
    import pandas as pd
    input_file.seek(0)
    warnings.filterwarnings(
        'ignore', 'Conditional Formatting', UserWarning, 'openpyxl'
    )
    res, plans = {'base': {}}, []
    ext = osp.splitext(input_file_name.lower())[1][1:]
    engine = xlref.Ref._engines.get(ext, xlref.Ref._engines[None])
    with pd.ExcelFile(io.BytesIO(input_file.read()), engine=engine) as xl:
        parent = Rererence('#A1')
        parent.ref['fpath'] = osp.abspath(input_file_name)
        parent.cache[parent.ref['fpath']] = parent.ref['xl_book'] = xl
        xl.sheet_indices = {k.lower(): i for i, k in enumerate(xl.sheet_names)}
        for sheet_name in xl.sheet_names:
            match = _re_input_sheet_name.match(sheet_name.strip(' '))
            if not match:
                log.debug("Sheet name '%s' cannot be parsed!", sheet_name)
                continue
            match = {k: v.lower() for k, v in match.groupdict().items() if v}
            is_plan = match.get('scope', None) == 'plan'
            if is_plan:
                r = {'plan': pd.DataFrame()}
            else:
                r = {}
            r = _parse_sheet(match, parent, sheet_name, res=r)
            if is_plan:
                plans.append(r['plan'])
            else:
                _add_times_base(r, **match)
                sh.combine_nested_dicts(r, depth=5, base=res)

    for k, v in sh.stack_nested_keys(res['base'], depth=3):
        if k[0] != 'target':
            v['cycle_type'] = v.get('cycle_type', k[-1].split('_')[0]).upper()
            v['cycle_name'] = v.get('cycle_name', k[-1]).upper()

    res['plan'] = _finalize_plan(res, plans, input_file_name).to_dict('records')

    return res



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
r"""
Define SYNCING the command line interface.

.. click:: co2mpas.cli.sync:cli
   :prog: co2mpas syncing
   :show-nested:

"""
import os
import click
import click_log
import os.path as osp
import schedula as sh
from syncing.cli import cli, logger

_sync_params = {p.name: p for p in cli.commands['sync'].params}
_sync_params['reference_name'].default = 'theoretical'
_sync_params['x_label'].default = ('times',)
_sync_params['y_label'].default = ('velocities',)


@cli.command('template', short_help='Generates sample template file.')
@click.argument(
    'output-file', default='datasync.xlsx', required=False,
    type=click.Path(writable=True)
)
@click.option(
    '-CT', '--cycle-type', default='wltp', show_default=True,
    type=click.Choice(['nedc', 'wltp']), help='Cycle type.'
)
@click.option(
    '-WC', '--wltp-class', default='class3b', show_default=True,
    type=click.Choice(['class1', 'class2', 'class3a', 'class3b']),
    help='WLTP vehicle class.'
)
@click.option(
    '-GB', '--gear-box-type', default='automatic', show_default=True,
    type=click.Choice(['manual', 'automatic']), help='Gear box type.'
)
@click_log.simple_verbosity_option(logger)
def template(output_file, cycle_type, gear_box_type, wltp_class):
    """
    Writes a sample template OUTPUT_FILE.

    OUTPUT_FILE: SYNCING input template file (.xlsx). [default: ./datasync.xlsx]
    """

    import pandas as pd
    from co2mpas.core.model.physical.cycle import dsp
    theoretical = sh.selector(['times', 'velocities'], dsp(inputs=dict(
        cycle_type=cycle_type.upper(), gear_box_type=gear_box_type,
        wltp_class=wltp_class, downscale_factor=0
    ), outputs=['times', 'velocities'], shrink=True))
    base = dict.fromkeys((
        'times', 'velocities', 'target gears', 'engine_speeds_out',
        'engine_coolant_temperatures', 'co2_normalization_references',
        'alternator_currents', 'battery_currents', 'target fuel_consumptions',
        'target co2_emissions', 'target engine_powers_out'
    ), [])
    data = dict(theoretical=theoretical, dyno=base, obd=base)
    os.makedirs(osp.dirname(output_file) or '.', exist_ok=True)
    with pd.ExcelWriter(output_file) as writer:
        for k, v in data.items():
            pd.DataFrame(v).to_excel(writer, k, index=False)
    return data


if __name__ == '__main__':
    cli()



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions to model the idle engine speed.
"""
import numpy as np
from sklearn.cluster import DBSCAN


# noinspection PyPep8Naming,PyMissingOrEmptyDocstring
class _IdleDetector(DBSCAN):
    def __init__(self, eps=0.5, min_samples=5, metric='euclidean',
                 algorithm='auto', leaf_size=30, p=None):
        super(_IdleDetector, self).__init__(
            eps=eps, min_samples=min_samples, metric=metric,
            algorithm=algorithm, leaf_size=leaf_size, p=p
        )
        self.cluster_centers_ = None
        self.min, self.max = None, None

    def fit(self, X, y=None, sample_weight=None):
        super(_IdleDetector, self).fit(X, y=y, sample_weight=sample_weight)

        c, lb = self.components_, self.labels_[self.core_sample_indices_]
        self.cluster_centers_ = np.array(
            [np.mean(c[lb == i]) for i in range(lb.max() + 1)]
        )
        self.min, self.max = c.min(), c.max()
        return self

    def predict(self, X, set_outliers=True):
        import sklearn.metrics as sk_met
        y = sk_met.pairwise_distances_argmin(X, self.cluster_centers_[:, None])
        if set_outliers:
            y[((X > self.max) | (X < self.min))[:, 0]] = -1
        return y
